"""Current vector control methods for synchronous machine drives."""

from dataclasses import dataclass, InitVar
import numpy as np
from motulator._helpers import abc2complex
from motulator.control._common import Ctrl, ComplexPICtrl, PWM, SpeedCtrl
from motulator.control.sm._torque import TorqueCharacteristics
from motulator.control.sm._observers import Observer
from motulator._utils import Bunch


# %%
@dataclass
class ModelPars:
    """
    Model parameters of a synchronous machine.
    
    Parameters
    ----------
    R_s : float
        Stator resistance (Ω).
    L_d : float
        d-axis inductance (H).
    L_q : float
        q-axis inductance (H).
    psi_f : float
        PM flux linkage (Vs).
    n_p : int
        Number of pole pairs.
    J : float
        Moment of inertia (kgm²).

    """
    R_s: float = None
    L_d: float = None
    L_q: float = None
    psi_f: float = None
    n_p: int = None
    J: float = None


# %%
class VectorCtrl(Ctrl):
    """
    Vector control for synchronous machine drives.

    This class interconnects the subsystems of the control system and
    provides the interface to the solver.

    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    ref : ReferencePars
        Reference generation parameters.
    T_s : float, optional
        Sampling period (s). The default is 250e-6.
    sensorless : bool, optional
        If True, sensorless control is used. The default is True.

    Attributes
    ----------
    current_ref : CurrentReference
        Current reference generator.
    observer : Observer
        Flux and rotor position observer, used in the sensorless mode only.
    current_ctrl : CurrentCtrl
        Current controller.
    speed_ctrl : SpeedCtrl 
        Speed controller.
    pwm : PWM
        Pulse-width modulation.
    w_m_ref : callable
        Speed reference (electrical rad/s) as a function of time (s).
        
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, par, ref, T_s=250e-6, sensorless=True):
        super().__init__()
        self.T_s = T_s
        self.sensorless = sensorless
        self.n_p = par.n_p
        self.current_ref = CurrentReference(par, ref)
        self.current_ctrl = CurrentCtrl(par, 2*np.pi*200)
        if sensorless:
            self.observer = Observer(par, alpha_o=2*np.pi*100, sensorless=True)
        else:
            self.observer = None
        self.speed_ctrl = SpeedCtrl(par.J, 2*np.pi*4)
        self.pwm = PWM()
        self.w_m_ref = callable

    # pylint: disable=too-many-locals
    def __call__(self, mdl):
        """
        Run the main control loop.

        Parameters
        ----------
        mdl : Drive
            Continuous-time system model for getting the feedback signals.

        Returns
        -------
        T_s : float
            Sampling period (s).
        d_abc : ndarray, shape (3,)
            Duty ratios.

        """
        # Get the speed reference
        w_m_ref = self.w_m_ref(self.clock.t)

        # Measure the feedback signals
        i_s_abc = mdl.machine.meas_currents()
        u_dc = mdl.converter.meas_dc_voltage()
        u_s = self.pwm.realized_voltage

        # Get the states
        if self.sensorless:
            w_m, theta_m = self.observer.w_m, self.observer.theta_m
        else:
            w_m = self.n_p*mdl.mechanics.meas_speed()
            theta_m = self.n_p*mdl.mechanics.meas_position()
            theta_m = np.mod(theta_m + np.pi, 2*np.pi) - np.pi

        # Current vector in estimated rotor coordinates
        i_s = np.exp(-1j*theta_m)*abc2complex(i_s_abc)

        # Outputs
        tau_M_ref = self.speed_ctrl.output(w_m_ref/self.n_p, w_m/self.n_p)
        i_s_ref, tau_M_ref_lim = self.current_ref.output(tau_M_ref, w_m, u_dc)
        u_s_ref = self.current_ctrl.output(i_s_ref, i_s)

        # Data logging
        data = Bunch(
            i_s=i_s,
            i_s_ref=i_s_ref,
            t=self.clock.t,
            tau_M_ref_lim=tau_M_ref_lim,
            theta_m=theta_m,
            u_dc=u_dc,
            u_s=u_s,
            w_m=w_m,
            w_m_ref=w_m_ref,
        )
        self.save(data)

        # Update states
        if self.sensorless:
            self.observer.update(self.T_s, u_s, i_s)
        self.speed_ctrl.update(self.T_s, tau_M_ref_lim)
        self.current_ref.update(self.T_s, tau_M_ref_lim, u_s_ref, u_dc)
        self.current_ctrl.update(self.T_s, u_s, w_m)
        self.clock.update(self.T_s)

        d_abc = self.pwm(self.T_s, u_s_ref, u_dc, theta_m, w_m)

        return self.T_s, d_abc


# %%
class CurrentCtrl(ComplexPICtrl):
    """
    Current controller for synchronous machines.

    This provides an interface of a current controller for synchronous machines
    [#Awa2019]_. The gains are initialized based on the desired closed-loop 
    bandwidth and the inductances. 

    Parameters
    ----------
    par : ModelPars
        Synchronous machine parameters, should contain `L_d` and `L_q` (H). 
    alpha_c : float
        Closed-loop bandwidth (rad/s).

    References
    ----------
    .. [#Awa2019] Awan, Saarakkala, Hinkkanen, "Flux-linkage-based current 
       control of saturated synchronous motors," IEEE Trans. Ind. Appl. 2019,
       https://doi.org/10.1109/TIA.2019.2919258

    """

    def __init__(self, par, alpha_c):
        k_p, k_i, k_t = 2*alpha_c, alpha_c**2, alpha_c
        super().__init__(k_p, k_i, k_t)
        self.L_d = par.L_d
        self.L_q = par.L_q

    def output(self, i_ref, i):
        # Extends the base class method by transforming the currents to the
        # flux linkages, which is a simple way to take the saliency into account
        psi_ref = self.L_d*i_ref.real + 1j*self.L_q*i_ref.imag
        psi = self.L_d*i.real + 1j*self.L_q*i.imag
        return super().output(psi_ref, psi)


# %%
# pylint: disable=too-many-instance-attributes
@dataclass
class CurrentReferencePars:
    """
    Parameters for reference generation.

    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    i_s_max : float
        Maximum stator current (A). 
    psi_s_min : float, optional
        Minimum stator flux (Vs). The default is `psi_f`.
    w_m_nom : float, optional
        Nominal rotor angular speed (electrical rad/s). Needed if `k_fw` is not
        directly provided.
    alpha_fw : float, optional
        Field-weakening bandwidth (rad/s). The default is 2*pi*20.
    k_fw : float, optional
        Field-weakening gain. The default is `alpha_fw/(w_m_nom*par.L_d)`.
    k_u : float, optional
        Voltage utilization factor. The default is 0.95.

    Attributes
    ----------
    i_sd_mtpa : callable
        MTPA d-axis current (A) as a funtion of the torque (Nm).
    tau_M_lim : callable
        Torque limite (Nm) as a function of the stator flux linkage (Vs). This
        limit merges the MTPV and current limits.
    i_sd_lim : callable
        d-axis current limit (A) as a function of the stator flux linkage (Vs).
        This limit merges the MTPV and current limits.
    
    """
    par: InitVar[ModelPars]
    i_s_max: float
    psi_s_min: float = None
    w_m_nom: InitVar[float] = None
    alpha_fw: InitVar[float] = 2*np.pi*20
    k_fw: float = None
    k_u: float = 0.95

    def __post_init__(self, par, w_m_nom, alpha_fw):
        # Minimum stator flux
        if self.psi_s_min is None:
            self.psi_s_min = par.psi_f
        # Field-weakening gain
        if self.k_fw is None:
            self.k_fw = alpha_fw/(w_m_nom*par.L_d)
        # Generate LUTs
        tq = TorqueCharacteristics(par)
        mtpa = tq.mtpa_locus(self.i_s_max, self.psi_s_min)
        lim = tq.mtpv_and_current_limits(self.i_s_max)
        # MTPA locus
        self.i_sd_mtpa = mtpa.i_sd_vs_tau_M
        # Merged MTPV and current limits
        self.tau_M_lim = lim.tau_M_vs_abs_psi_s
        self.i_sd_lim = lim.i_sd_vs_tau_M


# %%
class CurrentReference:
    """
    Current reference calculation.

    This method includes the MTPA locus and field-weakening operation based on
    the unlimited voltage reference feedback. The MTPV and current limits are
    taken into account. This resembles the method presented [#Bed2020]_.

    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    ref : CurrentReferencePars
        Reference generation parameters.

    Notes
    -----
    Instead of the PI controller used in [#Bed2020]_, we use a simpler integral
    controller with a constant gain. The resulting operating-point-dependent
    closed-loop pole could be derived using (12) of the paper. Unlike in 
    [#Bed2020]_, the MTPV limit is also included here by means of limiting the 
    reference torque and the d-axis current reference.

    References
    ----------
    .. [#Bed2020] Bedetti, Calligaro, Petrella, "Analytical design and 
       autotuning of adaptive flux-weakening voltage regulation loop in IPMSM 
       drives with accurate torque regulation," IEEE Trans. Ind. Appl., 2020,
       https://doi.org/10.1109/TIA.2019.2942807

    """

    # pylint: disable=too-many-instance-attributes, too-many-arguments
    def __init__(self, par, ref):
        # Machine model parameters
        self.n_p = par.n_p
        self.L_d, self.L_q, self.psi_f = par.L_d, par.L_q, par.psi_f
        # Reference generation parameters
        self.i_sd_mtpa = ref.i_sd_mtpa  # MTPA locus
        self.tau_M_lim = ref.tau_M_lim  # Merged MTPV and current limits
        self.i_sd_lim = ref.i_sd_lim  # Merged MTPV and current limits
        self.psi_s_min = ref.psi_s_min  # Minimum flux linkage
        self.i_s_max = ref.i_s_max  # Maximum current
        self.k_fw = ref.k_fw  # Field-weakening gain
        self.k_u = ref.k_u  # Voltage utilization factor
        # State
        self.i_sd_ref = 0

    def output(self, tau_M_ref, w_m, u_dc):
        """
        Compute the stator current reference.

        Parameters
        ----------
        tau_M_ref : float
            Torque reference (Nm).
        w_m : float
            Rotor speed (electrical rad/s)
        u_dc : float
            DC-bus voltage (V).

        Returns
        -------
        i_s_ref : complex
            Stator current reference (A).
        tau_M_ref_lim : float
            Limited torque reference (Nm).

        """

        def limit_torque(tau_M_ref, w_m, u_dc):
            if np.abs(w_m) > 0:
                psi_s_max = self.k_u*u_dc/np.sqrt(3)/np.abs(w_m)
                tau_M_max = self.tau_M_lim(psi_s_max)
            else:
                tau_M_max = self.tau_M_lim(np.inf)

            if np.abs(tau_M_ref) > tau_M_max:
                tau_M_ref = np.sign(tau_M_ref)*tau_M_max

            return tau_M_ref

        # Limit the torque reference according to MTPV and current limits
        tau_M_ref = limit_torque(tau_M_ref, w_m, u_dc)

        # q-axis current reference
        psi_t = self.psi_f + (self.L_d - self.L_q)*self.i_sd_ref
        i_sq_ref = tau_M_ref/(1.5*self.n_p*psi_t) if psi_t != 0 else 0

        # Limit the q-axis current reference
        i_sd_mtpa = self.i_sd_mtpa(np.abs(tau_M_ref))
        i_sq_max = np.min([
            np.sqrt(self.i_s_max**2 - self.i_sd_ref**2),
            np.sqrt(self.i_s_max**2 - i_sd_mtpa**2)
        ])
        if np.abs(i_sq_ref) > i_sq_max:
            i_sq_ref = np.sign(i_sq_ref)*i_sq_max

        # Current reference
        i_s_ref = self.i_sd_ref + 1j*i_sq_ref

        # Limited torque (for the speed controller)
        tau_M_ref_lim = 1.5*self.n_p*psi_t*i_sq_ref

        return i_s_ref, tau_M_ref_lim

    def update(self, T_s, tau_M_ref_lim, u_s_ref, u_dc):
        """
        Field-weakening based on the unlimited reference voltage.

        Parameters
        ----------
        T_s : float
            Sampling period (s).
        tau_M_ref_lim : float
            Limited torque reference (Nm).
        u_s_ref : complex
            Unlimited stator voltage reference (V).
        u_dc : float 
            DC-bus voltage (V).

        """
        u_s_max = self.k_u*u_dc/np.sqrt(3)
        self.i_sd_ref += T_s*self.k_fw*(u_s_max - np.abs(u_s_ref))

        # Limit the current
        i_sd_mtpa = self.i_sd_mtpa(np.abs(tau_M_ref_lim))
        i_sd_lim = self.i_sd_lim(np.abs(tau_M_ref_lim))
        self.i_sd_ref = np.clip(self.i_sd_ref, i_sd_lim, i_sd_mtpa)



"""
State observers for induction machines.
"""

import numpy as np


# %%
class Observer:
    """
    Reduced-order flux observer for induction machines in estimated 
    rotor flux coordinates.

    This class implements a reduced-order flux observer for induction machines.
    Both sensored and sensorless operation are supported. The observer structure
    is similar to [#Hin2010]_. The observer operates in estimated rotor flux coordinates. 
    
    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    k : callable, optional
        Observer gain as a function of the rotor angular speed. The default is 
        ``lambda w_m: (0.5*R_R/L_M + 0.2*abs(w_m))/(R_R/L_M - 1j*w_m)`` if
        `sensorless` else ``lambda w_m: 1 + 0.2*abs(w_m)/(R_R/L_M - 1j*w_m)``.
    alpha_o : float, optional
        Observer bandwidth (rad/s). The default is 2*pi*40.
    sensorless : bool, optional
        If True, sensorless mode is used. The default is True.

    Attributes
    ----------
    psi_R : float
        Rotor flux magnitude estimate (Vs).
    theta_s : float
        Rotor flux angle estimate (rad).
    w_m : float
        Rotor angular speed estimate (electrical rad/s). In sensored mode, this
        is the measured low-pass-filtered speed.

    Notes
    -----
    The pure voltage model corresponds to ``k = lambda w_m: 0``, resulting in 
    the marginally stable estimation-error dynamics. The current model is 
    obtained by setting ``k = lambda w_m: 1``. 

    References
    ----------
    .. [#Hin2010] Hinkkanen, Harnefors, Luomi, "Reduced-order flux observers 
       with stator-resistance adaptation for speed-sensorless induction motor 
       drives," IEEE Trans. Power Electron., 2010, 
       https://doi.org/10.1109/TPEL.2009.2039650

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, par, k=None, alpha_o=2*np.pi*40, sensorless=True):
        # Model parameters
        self.R_s, self.R_R, self.L_sgm = par.R_s, par.R_R, par.L_sgm
        self.alpha = par.R_R/par.L_M

        # Other parameters
        self.alpha_o = alpha_o
        self.sensorless = sensorless

        # Observer gains
        self.k1 = k
        if self.sensorless:
            if self.k1 is None:  # If not given, use the default gains
                self.k1 = lambda w_m: (.5*self.alpha + .2*np.abs(w_m))/(
                    self.alpha - 1j*w_m)
            self.k2 = self.k1
        else:
            if self.k1 is None:
                self.k1 = lambda w_m: 1 + .2*np.abs(w_m)/(self.alpha - 1j*w_m)
            self.k2 = 0*self.k1

        # States
        self.psi_R, self.theta_s, self.w_m, self._i_s_old = 0, 0, 0, 0

    def _f(self, T_s, u_s, i_s, w_m):
        # Right-hand-side of the differential equation.

        # Induced voltage from stator quantities (without the w_s term that is
        # taken into account separately to avoid an algebraic loop)
        v_s = u_s - self.R_s*i_s - self.L_sgm*(i_s - self._i_s_old)/T_s

        # Induced voltage from the rotor quantities
        v_r = self.R_R*i_s - (self.alpha - 1j*w_m)*self.psi_R

        # Angular frequency of the rotor flux vector
        k1, k2 = self.k1(w_m), self.k2(w_m)
        den = self.psi_R + self.L_sgm*np.real((1 - k1)*i_s + k2*np.conj(i_s))
        num = np.imag(v_s + k1*(v_r - v_s) + k2*np.conj(v_r - v_s))
        w_s = num/den if den > 0 else w_m

        # Compute the derivative of the flux magnitude for the state update
        v = v_s - 1j*w_s*self.L_sgm*i_s
        dpsi_R = np.real(v + k1*(v_r - v) + k2*np.conj(v_r - v))

        return dpsi_R, w_s  # Note: w_s = dtheta_s

    # pylint: disable=too-many-arguments
    def _update(self, T_s, i_s, dpsi_R, w_s, w_m):
        # Update the states
        self.psi_R += T_s*dpsi_R
        self.theta_s += T_s*w_s  # Next line: limit into [-pi, pi)
        self.theta_s = np.mod(self.theta_s + np.pi, 2*np.pi) - np.pi
        self.w_m += T_s*self.alpha_o*(w_m - self.w_m)
        # Store the old current
        self._i_s_old = i_s

    def __call__(self, T_s, u_s, i_s, w_m=None):
        """
        Compute the angular frequency of the flux and update the states.

        Parameters
        ----------
        T_s : float
            Sampling period (s).
        u_s : complex
            Stator voltage (V) in estimated rotor flux coordinates.
        i_s : complex
            Stator current (A) in estimated rotor flux coordinates.
        w_m : float, optional
            Rotor angular speed (electrical rad/s). This signal is only used in
            sensored mode. The default is None.

        Returns
        -------
        w_s : float
            Angular frequency (rad/s) of the rotor flux estimate.

        """
        # Compute the time derivative of the flux estimate
        if self.sensorless:
            dpsi_R, w_s = self._f(T_s, u_s, i_s, self.w_m)
            # Slip and unfiltered speed estimates
            w_r = self.R_R*i_s.imag/self.psi_R if self.psi_R > 0 else 0
            w_m = w_s - w_r
        else:
            dpsi_R, w_s = self._f(T_s, u_s, i_s, w_m)

        # Update the states
        self._update(T_s, i_s, dpsi_R, w_s, w_m)

        return w_s


class FullOrderObserver:
    """
    Full-order flux observer for induction machines in estimated 
    rotor flux coordinates.

    This class implements a full-order flux observer for induction machines.
    The observer structure is similar to [#Tii2023]_. The observer operates in 
    estimated rotor flux coordinates. 
    
    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    k : callable, optional
        Observer gain as a function of the rotor angular speed. The default is 
        ``lambda w_m: (R_R/L_M + 0.2*abs(w_m))/(R_R/L_M - 1j*w_m)`` 
    alpha_o : float, optional
        Observer bandwidth (rad/s). The default is 2*pi*40.
    alpha_i : float, optional
        Current estimation bandwidth (rad/s). The default is 2*pi*400.
    sensorless : bool, optional
        If True, sensorless mode is used. The default is True.

    Attributes
    ----------
    psi_R : float
        Rotor flux magnitude estimate (Vs).
    i_s : float
        Stator current estimate (A).
    theta_s : float
        Rotor flux angle estimate (rad).
    w_m : float
        Integral state of the rotor angular speed estimate (electrical rad/s). 


    References
    ----------
    .. [#Tii2023] Tiitinen, Hinkkanen, Harnefors, "Speed-Adaptive Full-Order Observer 
        Revisited: Closed-Form Design for Induction Motor Drives," Proc. IEEE SLED, 
        Seoul, South Korea,  Aug. 2023, https://doi.org/10.1109/SLED57582.2023.10261359

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(
            self,
            par,
            k=None,
            alpha_o=2*np.pi*40,
            alpha_i=2*np.pi*400,
            sensorless=True):
        # Model parameters
        self.R_s, self.R_R, self.L_sgm = par.R_s, par.R_R, par.L_sgm
        self.alpha = par.R_R/par.L_M

        # Other parameters
        self.alpha_o = alpha_o
        self.alpha_i = alpha_i
        self.g = alpha_i - self.alpha
        self.sensorless = sensorless

        # Observer gains
        self.k = k
        if self.sensorless:
            if self.k is None:  # If not given, use the default gains
                self.k = lambda w_m: (self.alpha + .2*np.abs(w_m))/(
                    self.alpha - 1j*w_m)
        else:
            raise NotImplementedError(
                'Sensored mode not implemented for full-order observer')

        # States
        self.psi_R, self.i_s, self.theta_s, self.w_m = 0, 0, 0, 0

    def _f(self, u_s, i_s, w_m):

        # Current estimation error
        i_err = i_s - self.i_s

        # Angular frequency of the rotor flux vector
        den = self.psi_R - self.L_sgm*i_err.real
        num = self.R_R*i_s.imag + self.L_sgm*(
            self.alpha_i*self.k(w_m).imag*i_err.real - self.g*i_err.imag)
        w_s = w_m + num/den if den > 0 else w_m

        # Compute the derivative of the current (vector) estimate
        # and flux magnitude for the state update.
        k_i = self.L_sgm*(self.g - 1j*(w_s - w_m)) - (self.R_R + self.R_s)
        di_s = (
            u_s - (self.R_s + self.R_R + 1j*w_s*self.L_sgm)*self.i_s +
            (self.alpha - 1j*w_m)*self.psi_R + k_i*i_err)/self.L_sgm

        dpsi_R = (
            -self.alpha*self.psi_R + self.R_R*i_s.real +
            (self.alpha_i*self.k(w_m).real - self.g)*self.L_sgm*i_err.real -
            (w_s - w_m)*self.L_sgm*i_err.imag)

        # Derivative of the integral term of the speed estimate
        dw_m = -self.alpha_o*self.alpha_i*self.L_sgm*i_err.imag/self.psi_R if self.psi_R > 0 else 0

        return di_s, dpsi_R, dw_m, w_s

    # pylint: disable=too-many-arguments
    def _update(self, T_s, di_s, dpsi_R, dw_m, w_s):
        # Update the states
        self.i_s += T_s*di_s
        self.psi_R += T_s*dpsi_R
        self.w_m += T_s*dw_m
        self.theta_s += T_s*w_s  # Next line: limit into [-pi, pi)
        self.theta_s = np.mod(self.theta_s + np.pi, 2*np.pi) - np.pi

    def __call__(self, T_s, u_s, i_s, w_m=None):
        """
        Compute the angular frequency of the flux and update the states.

        Parameters
        ----------
        T_s : float
            Sampling period (s).
        u_s : complex
            Stator voltage (V) in estimated rotor flux coordinates.
        i_s : complex
            Stator current (A) in estimated rotor flux coordinates.
        w_m : float, optional
            Rotor angular speed (electrical rad/s). This signal is not used
            and only exists for compatibility.

        Returns
        -------
        w_s : float
            Angular frequency (rad/s) of the rotor flux estimate.

        """
        # PI-type speed adaptation
        p_term = -self.alpha_o*self.L_sgm*(
            i_s - self.i_s).imag/self.psi_R if self.psi_R > 0 else 0
        w_m = p_term + self.w_m

        # Compute the time derivative of the flux estimate
        di_s, dpsi_R, dw_m, w_s = self._f(u_s, i_s, w_m)

        # Update the states
        self._update(T_s, di_s, dpsi_R, dw_m, w_s)

        return w_s


# %%
class FluxObserver:
    """
    Sensorless reduced-order flux observer in external coordinates.

    This is a sensorless reduced-order flux observer in synchronous coordinates
    for an induction machine. The observer gain decouples the electrical and
    mechanical dynamics and allows placing the poles of the linearized 
    estimation error dynamics. This implementation operates in external 
    coordinates (typically synchronous coordinates defined by reference signals 
    of a control system).

    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    alpha_o : float
        Speed-estimation bandwidth (rad/s).
    b : callable, optional 
        Coefficient (rad/s) of the characteristic polynomial as a function of 
        the rotor angular speed estimate. The default is 
        ``lambda w_m: R_R/L_M + .4*abs(w_m)``.

    Attributes
    ----------
    psi_R : complex
        Rotor flux estimate (Vs).
    w_m : float
        Rotor angular speed estimate (electrical rad/s).

    Notes
    -----
    The characteristic polynomial of the observer in synchronous coordinates is 
    ``s**2 + b*s + w_s**2``.  

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, par, alpha_o, b=None):
        self.R_s, self.R_R, self.L_sgm = par.R_s, par.R_R, par.L_sgm
        self.alpha = par.R_R/par.L_M
        # Design parameters
        self.alpha_o = alpha_o
        zeta_inf = .2  # Damping ratio at high speeds
        self.b = lambda w_m: (
            self.alpha + 2*zeta_inf*np.abs(w_m) if b is None else b)
        # States
        self.psi_R, self.w_m = 0, 0
        # Previous current, private attribute
        self._i_s_old = 0

    def update(self, T_s, u_s, i_s, w_s):
        """
        Update the states.

        Parameters
        ----------
        T_s : float
            Sampling period (s).
        u_s : complex
            Stator voltage (V) in synchronous coordinates.
        i_s : complex
            Stator current (A) in synchronous coordinates.
        w_s : float
            Angular frequency (rad/s) of the coordinate system. 

        """
        # Observer gain with c = w_s**2 (without the orthogonal projection
        # which is embedded into the error signal and the state update)
        g = self.b(self.w_m)/(self.alpha - 1j*self.w_m)

        # Time derivative of the stator current
        di_s = (i_s - self._i_s_old)/T_s

        # Induced voltage from the rotor quantities
        e_r = self.R_R*i_s - (self.alpha - 1j*self.w_m)*self.psi_R

        # Induced voltage from stator quantities
        e_s = u_s - self.R_s*i_s - self.L_sgm*(di_s + 1j*w_s*i_s)

        # Error signal (rad/s)
        err = (e_s - e_r)/self.psi_R if np.abs(self.psi_R) > 0 else 0

        # Update the states
        self.psi_R += T_s*(e_s - (1j*w_s + g*err.real)*self.psi_R)
        self.w_m += T_s*self.alpha_o*err.imag
        self._i_s_old = i_s



"""Flux-vector control of synchronous machine drives."""

from dataclasses import dataclass, InitVar
import numpy as np
from motulator._helpers import abc2complex
from motulator.control._common import Ctrl, PWM, SpeedCtrl
from motulator.control.sm._torque import TorqueCharacteristics
from motulator.control.sm._vector import ModelPars
from motulator.control.sm._observers import Observer
from motulator._utils import Bunch


# %%
class FluxVectorCtrl(Ctrl):
    """
    Flux-vector control of synchronous machine drives.

    This class implements a variant of stator-flux-vector control [#Pel2009]_. 
    Rotor coordinates as well as decoupling between the stator flux and torque 
    channels are used according to [#Awa2019]_. Here, the stator flux magnitude 
    and the electromagnetic torque are selected as controllable variables. 

    Notes
    -----
    Proportional controllers are used for simplicity. The magnetic saturation is 
    not considered in this implementation.

    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    ref : FluxTorqueReferencePars
        Reference generation parameters.
    alpha_psi : float, optional
        Bandwidth of the flux controller (rad/s). The default is 2*pi*100.
    alpha_tau : float, optional
        Bandwidth of the torque controller (rad/s). The default is 2*pi*200.
    T_s : float
        Sampling period (s). The default is 250e-6.
    sensorless : bool, optional
        If True, sensorless control is used. The default is True.

    Attributes
    ----------
    observer : Observer
        Flux observer, having both sensorless and sensored modes.
    flux_torque_ref : FluxTorqueReference
        Flux and torque reference generator.
    speed_ctrl : SpeedCtrl
        Speed controller.
    w_m_ref : float
        Speed reference (electrical rad/s) as a function of time (s).
    pwm : PWM
        Pulse-width modulation.

    References
    ----------
    .. [#Pel2009] Pellegrino, Armando, Guglielmi, “Direct flux field-oriented 
       control of IPM drives with variable DC link in the field-weakening 
       region,” IEEE Trans.Ind. Appl., 2009, 
       https://doi.org/10.1109/TIA.2009.2027167

    .. [#Awa2019] Awan, Hinkkanen, Bojoi, Pellegrino, "Stator-flux-oriented 
       control of synchronous motors: A systematic design procedure," IEEE Trans. 
       Ind. Appl., 2019, https://doi.org/10.1109/TIA.2019.2927316

    """

    # pylint: disable=too-many-instance-attributes, too-many-arguments
    def __init__(
            self,
            par,
            ref,
            alpha_psi=2*np.pi*100,
            alpha_tau=2*np.pi*200,
            T_s=250e-6,
            sensorless=True):
        super().__init__()
        self.T_s = T_s
        self.sensorless = sensorless
        # Machine model parameters
        self.n_p = par.n_p
        self.R_s, self.L_d, self.L_q = par.R_s, par.L_d, par.L_q
        # Subsystems
        self.observer = Observer(
            par, alpha_o=2*np.pi*100, sensorless=sensorless)
        self.flux_torque_ref = FluxTorqueReference(ref)
        self.pwm = PWM()
        self.speed_ctrl = SpeedCtrl(par.J, 2*np.pi*4)
        # Bandwidths
        self.alpha_psi = alpha_psi
        self.alpha_tau = alpha_tau
        # Speed reference
        self.w_m_ref = callable

    # pylint: disable=too-many-locals
    def __call__(self, mdl):
        """
        Run the main control loop.

        Parameters
        ----------
        mdl : Drive
            Continuous-time system model for getting the feedback signals.

        Returns
        -------
        T_s : float
            Sampling period (s).
        d_abc : ndarray, shape (3,)
            Duty ratios.

        """
        # Get the speed reference
        w_m_ref = self.w_m_ref(self.clock.t)

        # Feedback signals
        i_s_abc = mdl.machine.meas_currents()  # Phase currents
        u_dc = mdl.converter.meas_dc_voltage()  # DC-bus voltage
        u_s = self.pwm.realized_voltage  # Realized voltage from PWM
        #
        if self.sensorless:
            # Get the rotor speed and position estimates
            w_m, theta_m = self.observer.w_m, self.observer.theta_m
        else:
            # Measure the rotor speed
            w_m = self.n_p*mdl.mechanics.meas_speed()
            # Limit the electrical rotor position into [-pi, pi)
            theta_m = np.mod(
                self.n_p*mdl.mechanics.meas_position() + np.pi,
                2*np.pi) - np.pi

        # Current vector in estimated rotor coordinates
        i_s = np.exp(-1j*theta_m)*abc2complex(i_s_abc)

        # Flux and torque estimates
        psi_s = self.observer.psi_s
        tau_M = 1.5*self.n_p*np.imag(i_s*np.conj(psi_s))

        # Outputs
        tau_M_ref = self.speed_ctrl.output(w_m_ref/self.n_p, w_m/self.n_p)
        psi_s_ref, tau_M_ref_lim = self.flux_torque_ref(tau_M_ref, w_m, u_dc)

        # Auxiliary current
        i_a = psi_s.real/self.L_q + 1j*psi_s.imag/self.L_d - i_s

        # Torque-production factor (c_tau = 0 corresponds to the MTPV condition)
        c_tau = 1.5*self.n_p*np.real(i_a*np.conj(psi_s))

        # References for the flux and torque controllers
        v_psi = self.alpha_psi*(psi_s_ref - np.abs(psi_s))
        v_tau = self.alpha_tau*(tau_M_ref_lim - tau_M)
        if c_tau > 0:
            v = (1.5*self.n_p*np.abs(psi_s)*i_a*v_psi + 1j*psi_s*v_tau)/c_tau
        else:
            v = v_psi

        # Stator voltage reference
        u_s_ref = self.R_s*i_s + 1j*w_m*psi_s + v

        # Data logging
        data = Bunch(
            i_s=i_s,
            psi_s=psi_s,
            psi_s_ref=psi_s_ref,
            t=self.clock.t,
            tau_M_ref_lim=tau_M_ref_lim,
            theta_m=theta_m,
            u_dc=u_dc,
            u_s=u_s,
            w_m=w_m,
            w_m_ref=w_m_ref,
        )
        self.save(data)

        # Update states
        self.observer.update(self.T_s, u_s, i_s, w_m)
        self.speed_ctrl.update(self.T_s, tau_M_ref_lim)
        self.clock.update(self.T_s)

        # PWM output
        d_abc = self.pwm(self.T_s, u_s_ref, u_dc, theta_m, w_m)

        return self.T_s, d_abc


# %%
@dataclass
class FluxTorqueReferencePars:
    """
    Parameters for reference generation.

    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    i_s_max : float
        Maximum stator current (A). 
    psi_s_min : float, optional
        Minimum stator flux (Vs). The default is `psi_f`.
    psi_s_max : float, optional
        Maximum stator flux (Vs). The default is inf.
    k_u : float, optional
        Voltage utilization factor. The default is 0.95.

    Attributes
    ----------
    psi_s_mtpa : callable
        MTPA stator flux linkage (Vs) as a function of the torque (Nm).
    tau_M_lim : callable
        Torque limit (Nm) as a function of the stator flux linkage (Vs). This 
        limit merges the MTPV and current limits. 

    """
    par: InitVar[ModelPars]
    i_s_max: float = None
    psi_s_min: float = None
    psi_s_max: float = np.inf
    k_u: float = 0.95

    def __post_init__(self, par):
        self.psi_s_min = par.psi_f if self.psi_s_min is None else self.psi_s_min
        # Generate LUTs
        tq = TorqueCharacteristics(par)
        mtpa = tq.mtpa_locus(self.i_s_max, self.psi_s_min)
        lim = tq.mtpv_and_current_limits(self.i_s_max)
        # MTPA locus
        self.psi_s_mtpa = mtpa.abs_psi_s_vs_tau_M
        # Merged MTPV and current limits
        self.tau_M_lim = lim.tau_M_vs_abs_psi_s


# %%
class FluxTorqueReference:
    """
    Flux and torque references.

    The current and MTPV limits as well as the MTPA locus are implemented as
    look-up tables, which are generated based on the constant machine model
    parameters.

    Parameters
    ----------
    ref : FluxTorqueReferencePars
        Reference generation parameters.

    """

    # pylint: disable=too-many-arguments
    def __init__(self, ref):
        self.psi_s_min, self.psi_s_max = ref.psi_s_min, ref.psi_s_max
        self.k_u = ref.k_u
        # Merged MTPV and current limits
        self.tau_M_lim = ref.tau_M_lim
        # MTPA locus
        self.psi_s_mtpa = ref.psi_s_mtpa

    def __call__(self, tau_M_ref, w_m, u_dc):
        """
        Calculate the stator flux reference and limit the torque reference.

        Parameters
        ----------
        tau_M_ref : float
            Unlimited torque reference (Nm).
        w_m : float
            Rotor speed or its reference (electrical rad/s).
        u_dc : float
            DC-bus voltage (V).

        Returns
        -------
        psi_s_ref : float
            Stator flux reference (Vs).
        tau_M_ref_lim : float
            Limited torque reference (Nm).

        """
        # Get the MTPA flux
        psi_s_mtpa = self.psi_s_mtpa(np.abs(tau_M_ref))
        psi_s_mtpa = np.clip(psi_s_mtpa, self.psi_s_min, self.psi_s_max)

        # Field weakening
        u_s_max = self.k_u*u_dc/np.sqrt(3)
        psi_s_max = u_s_max/np.abs(w_m) if np.abs(w_m) > 0 else np.inf

        # Flux reference
        psi_s_ref = np.min([psi_s_max, psi_s_mtpa])

        # Limit the torque reference according to the MTPV and current limits
        tau_M_lim = self.tau_M_lim(psi_s_ref)
        tau_M_ref_lim = np.min([tau_M_lim, np.abs(tau_M_ref)
                                ])*np.sign(tau_M_ref)

        return psi_s_ref, tau_M_ref_lim



"""
Continuous-time models for synchronous motor drives.

Peak-valued complex space vectors are used.

"""
import numpy as np
from motulator._helpers import complex2abc
from motulator._utils import Bunch


# %%
class SynchronousMachine:
    """
    Synchronous machine model.

    This models a synchronous machine in rotor coordinates. The stator flux 
    linkage and the electrical angle of the rotor are the state variables. 

    Parameters
    ----------
    n_p : int
        Number of pole pairs.
    R_s : float
        Stator resistance (Ω).
    L_d : float
        d-axis inductance (H).
    L_q : float
        q-axis inductance (H).
    psi_f : float
        PM-flux linkage (Vs).

    """

    def __init__(self, n_p, R_s, L_d, L_q, psi_f):
        # pylint: disable=too-many-arguments
        self.n_p, self.R_s = n_p, R_s
        self.L_d, self.L_q, self.psi_f = L_d, L_q, psi_f
        # Initial values
        self.psi_s0 = complex(psi_f)
        self.theta_m0 = 0

    def current(self, psi_s):
        """
        Compute the stator current.

        Parameters
        ----------
        psi_s : complex
            Stator flux linkage (Vs).

        Returns
        -------
        i_s : complex
            Stator current (A).

        """
        i_s = (psi_s.real - self.psi_f)/self.L_d + 1j*psi_s.imag/self.L_q
        return i_s

    def magnetic(self, psi_s):
        """
        Magnetic model.

        Parameters
        ----------
        psi_s : complex
            Stator flux linkage (Vs).

        Returns
        -------
        i_s : complex
            Stator current (A).
        tau_M : float
            Electromagnetic torque (Nm).

        """
        i_s = self.current(psi_s)
        tau_M = 1.5*self.n_p*np.imag(i_s*np.conj(psi_s))
        return i_s, tau_M

    def f(self, psi_s, u_s, w_M):
        """
        Compute the state derivative.

        Parameters
        ----------
        psi_s : complex
            Stator flux linkage (Vs).
        u_s : complex
            Stator voltage (V).
        w_M : float
            Rotor angular speed (mechanical rad/s).

        Returns
        -------
        complex list, length 2
            Time derivative of the state vector, [dpsi_s, dtheta_m0]
        i_s : complex
            Stator current (A).
        tau_M : float
            Electromagnetic torque (Nm).

        Notes
        -----
        In addition to the state derivative, this method also returns the 
        output signals (stator current `i_ss` and torque `tau_M`) needed for 
        interconnection with other subsystems. This avoids overlapping 
        computation in simulation.

        """
        i_s, tau_M = self.magnetic(psi_s)
        dpsi_s = u_s - self.R_s*i_s - 1j*self.n_p*w_M*psi_s
        dtheta_m = self.n_p*w_M
        return [dpsi_s, dtheta_m], i_s, tau_M

    def meas_currents(self):
        """
        Measure the phase currents at the end of the sampling period.

        Returns
        -------
        i_s_abc : 3-tuple of floats
            Phase currents (A).

        """
        i_s0 = self.current(self.psi_s0)
        i_s_abc = complex2abc(np.exp(1j*self.theta_m0)*i_s0)
        return i_s_abc


# %%
class SynchronousMachineSaturated(SynchronousMachine):
    """
    Model of a saturated synchronous machine.

    This overrides the linear magnetics model of the SynchronousMachine class
    with a generic saturation model::

        i_s = i_s(psi_s)
        
    The saturation model could be an analytical function or a look-up table. 

    Parameters
    ----------
    n_p : int
        Number of pole pairs.
    R_s : float
        Stator resistance (Ω).
    current : callable
        Function that computes the stator current `i_s` as a function of the 
        stator flux linkage `psi_s`. 
    psi_s0 : complex, optional
        Initial value of the stator flux linkage (Vs). For PM machines, this 
        should be solved from the the saturation model. The default is 0j. 

    """

    def __init__(self, n_p, R_s, current, psi_s0=0j):
        # pylint: disable=super-init-not-called
        self.n_p, self.R_s = n_p, R_s
        self.current = current
        # Initial values
        self.psi_s0 = complex(psi_s0)
        self.theta_m0 = 0


# %%
class Drive:
    """
    Continuous-time model for a synchronous machine drive.

    This interconnects the subsystems of a synchronous machine drive and 
    provides an interface to the solver. More complicated systems could be 
    modeled using a similar template.

    Parameters
    ----------
    machine : SynchronousMachine
        Synchronous machine model.
    mechanics : Mechanics
        Mechanics model.
    converter : Inverter
        Inverter model.

    """

    def __init__(self, machine=None, mechanics=None, converter=None):
        self.machine = machine
        self.mechanics = mechanics
        self.converter = converter

        # Initial time
        self.t0 = 0

        # Store the solution in these lists
        self.data = Bunch()
        self.data.t, self.data.q = [], []
        self.data.psi_s, self.data.theta_m = [], []
        self.data.w_M, self.data.theta_M = [], []

    def get_initial_values(self):
        """
        Get the initial values.

        Returns
        -------
        x0 : complex list, length 4
            Initial values of the state variables.

        """
        x0 = [
            self.machine.psi_s0,
            self.machine.theta_m0,
            self.mechanics.w_M0,
            self.mechanics.theta_M0,
        ]
        return x0

    def set_initial_values(self, t0, x0):
        """
        Set the initial values.

        Parameters
        ----------
        t0 : float
            Initial time (s).
        x0 : complex ndarray
            Initial values of the state variables.

        """
        self.t0 = t0
        self.machine.psi_s0 = x0[0]
        # x0[1:3].imag are always zero
        self.machine.theta_m0 = x0[1].real
        self.mechanics.w_M0 = x0[2].real
        # Limit the angles into [-pi, pi)
        self.mechanics.theta_m0 = np.mod(x0[1].real + np.pi, 2*np.pi) - np.pi
        self.mechanics.theta_M0 = np.mod(x0[3].real + np.pi, 2*np.pi) - np.pi

    def f(self, t, x):
        """
        Compute the complete state derivative list for the solver.

        Parameters
        ----------
        t : float
            Time (s).
        x : complex ndarray
            State vector.

        Returns
        -------
        complex list
            State derivatives.

        """
        # Unpack the states
        psi_s, theta_m, w_M, _ = x

        # Interconnections: outputs for computing the state derivatives
        u_ss = self.converter.ac_voltage(
            self.converter.q, self.converter.u_dc0)
        u_s = np.exp(-1j*theta_m)*u_ss  # Stator voltage in rotor coordinates

        # State derivatives
        machine_f, _, tau_M = self.machine.f(psi_s, u_s, w_M)
        mechanics_f = self.mechanics.f(t, w_M, tau_M)

        # List of state derivatives
        return machine_f + mechanics_f

    def save(self, sol):
        """
        Save the solution.

        Parameters
        ----------
        sol : Bunch
            Solution from the solver.

        """
        self.data.t.extend(sol.t)
        self.data.q.extend(sol.q)
        self.data.psi_s.extend(sol.y[0])
        self.data.theta_m.extend(sol.y[1].real)
        self.data.w_M.extend(sol.y[2].real)
        self.data.theta_M.extend(sol.y[3].real)

    def post_process(self):
        """Transform the lists to the ndarray format and post-process them."""
        # From lists to the ndarray
        for key in self.data:
            self.data[key] = np.asarray(self.data[key])

        # Compute some useful quantities
        self.data.i_s, self.data.tau_M = self.machine.magnetic(self.data.psi_s)
        self.data.w_m = self.machine.n_p*self.data.w_M
        self.data.tau_L = (
            self.mechanics.tau_L_t(self.data.t) +
            self.mechanics.tau_L_w(self.data.w_M))
        self.data.u_ss = self.converter.ac_voltage(
            self.data.q, self.converter.u_dc0)
        self.data.theta_m = np.mod(  # Limit into [-pi, pi)
            self.data.theta_m + np.pi, 2*np.pi) - np.pi
        self.data.theta_M = np.mod(  # Limit into [-pi, pi)
            self.data.theta_M + np.pi, 2*np.pi) - np.pi
        self.data.i_ss = self.data.i_s*np.exp(1j*self.data.theta_m)


# %%
class DriveTwoMassMechanics(Drive):
    """
    Synchronous machine drive with two-mass mechanics.

    This interconnects the subsystems of a synchronous machine drive and
    provides an interface to the solver.

    Parameters
    ----------
    machine : SynchronousMachine
        Synchronous machine model.
    mechanics : MechanicsTwoMass
        Mechanics model.
    converter : Inverter
        Inverter model.

    """

    def __init__(self, machine=None, mechanics=None, converter=None):
        super().__init__(machine, mechanics, converter)
        self.data.w_L, self.data.theta_ML = [], []

    def get_initial_values(self):
        """Extend the base class."""
        x0 = super().get_initial_values() + [
            self.mechanics.w_L0, self.mechanics.theta_ML0
        ]
        return x0

    def set_initial_values(self, t0, x0):
        """Extend the base class."""
        super().set_initial_values(t0, x0[0:4])
        self.mechanics.w_L0 = x0[4].real
        self.mechanics.theta_ML0 = np.mod(x0[5].real + np.pi, 2*np.pi) - np.pi

    def f(self, t, x):
        """Override the base class."""
        # Unpack the states
        psi_s, theta_m, w_M, _, w_L, theta_ML = x
        # Interconnections: outputs for computing the state derivatives
        u_ss = self.converter.ac_voltage(
            self.converter.q, self.converter.u_dc0)
        u_s = np.exp(-1j*theta_m)*u_ss  # Stator voltage in rotor coordinates
        # State derivatives plus the outputs for interconnections
        machine_f, _, tau_M = self.machine.f(psi_s, u_s, w_M)
        mechanics_f = self.mechanics.f(t, w_M, w_L, theta_ML, tau_M)
        # List of state derivatives
        return machine_f + mechanics_f

    def save(self, sol):
        """Extend the base class."""
        super().save(sol)
        self.data.w_L.extend(sol.y[4].real)
        self.data.theta_ML.extend(sol.y[5].real)

    def post_process(self):
        """Extend the base class."""
        super().post_process()
        # From lists to the ndarray
        self.data.w_L = np.asarray(self.data.w_L)
        self.data.theta_ML = np.asarray(self.data.theta_ML)



"""Helper functions and classes."""

# %%
from dataclasses import dataclass
import numpy as np


# %%
def abc2complex(u):
    """
    Transform three-phase quantities to a complex space vector.

    Parameters
    ----------
    u : array_like, shape (3,)
        Phase quantities.

    Returns
    -------
    complex
        Complex space vector (peak-value scaling).

    Examples
    --------
    >>> from motulator import abc2complex
    >>> y = abc2complex([1, 2, 3])
    >>> y
    (-1-0.5773502691896258j)

    """
    return (2/3)*u[0] - (u[1] + u[2])/3 + 1j*(u[1] - u[2])/np.sqrt(3)


# %%
def complex2abc(u):
    """
    Transform a complex space vector to three-phase quantities.

    Parameters
    ----------
    u : complex
        Complex space vector (peak-value scaling).

    Returns
    -------
    ndarray, shape (3,)
        Phase quantities.

    Examples
    --------
    >>> from motulator import complex2abc
    >>> y = complex2abc(1-.5j)
    >>> y
    array([ 1.       , -0.9330127, -0.0669873])

    """
    return np.array([
        u.real, .5*(-u.real + np.sqrt(3)*u.imag),
        .5*(-u.real - np.sqrt(3)*u.imag)
    ])


# %%
@dataclass
class BaseValues:
    """
    Base values.

    Base values are computed from the nominal values and the number of pole
    pairs. They can be used, e.g., for scaling the plotted waveforms.

    Parameters
    ----------
    U_nom : float
        Voltage (V, rms, line-line).
    I_nom : float
        Current (A, rms).
    f_nom : float
        Frequency (Hz).
    tau_nom : float
        Torque (Nm).
    P_nom : float
        Power (W).
    n_p : int
        Number of pole pairs.

    Attributes
    ----------
    u : float
        Base voltage (V, peak, line-neutral).
    i : float
        Base current (A, peak).
    w : float
        Base angular frequency (rad/s).
    psi : float
        Base flux linkage (Vs).
    p : float
        Base power (W).
    Z : float
        Base impedance (Ω).
    L : float
        Base inductance (H).
    tau : float
        Base torque (Nm).

    """

    # pylint: disable=too-many-instance-attributes
    U_nom: float
    I_nom: float
    f_nom: float
    P_nom: float
    tau_nom: float
    n_p: int

    def __post_init__(self):
        self.u = np.sqrt(2/3)*self.U_nom
        self.i = np.sqrt(2)*self.I_nom
        self.w = 2*np.pi*self.f_nom
        self.psi = self.u/self.w
        self.p = 1.5*self.u*self.i
        self.Z = self.u/self.i
        self.L = self.Z/self.w
        self.tau = self.n_p*self.p/self.w


# %%
class Sequence:
    """
    Sequence generator.

    The time array must be increasing. The output values are interpolated
    between the data points.

    Parameters
    ----------
    times : ndarray
        Time values.
    values : ndarray
        Output values.
    periodic : bool, optional
        Enables periodicity. The default is False.

    """

    def __init__(self, times, values, periodic=False):
        self.times = times
        self.values = values
        if periodic is True:
            self._period = times[-1] - times[0]
        else:
            self._period = None

    def __call__(self, t):
        """
        Interpolate the output.

        Parameters
        ----------
        t : float
            Time.

        Returns
        -------
        float or complex
            Interpolated output.

        """
        return np.interp(t, self.times, self.values, period=self._period)


# %%
class Step:
    """Step function."""

    def __init__(self, step_time, step_value, initial_value=0):
        self.step_time = step_time
        self.step_value = step_value
        self.initial_value = initial_value

    def __call__(self, t):
        """
        Step function.

        Parameters
        ----------
        t : float
            Time.

        Returns
        -------
        float
            Step output.

        """
        return self.initial_value + (t >= self.step_time)*self.step_value



"""Continuous-time induction machine models."""

from motulator.model.im._drive import (
    Drive,
    DriveWithDiodeBridge,
    DriveTwoMassMechanics,
    InductionMachine,
    InductionMachineSaturated,
    InductionMachineInvGamma,
)

__all__ = [
    "Drive",
    "DriveWithDiodeBridge",
    "DriveTwoMassMechanics",
    "InductionMachine",
    "InductionMachineSaturated",
    "InductionMachineInvGamma",
]



"""Continuous-time system models."""

from motulator.model._mechanics import Mechanics, MechanicsTwoMass
from motulator.model._converter import FrequencyConverter, Inverter
from motulator.model._simulation import CarrierComparison, Simulation

import motulator.model.im as im
import motulator.model.sm as sm

__all__ = [
    "Mechanics",
    "MechanicsTwoMass",
    "FrequencyConverter",
    "Inverter",
    "CarrierComparison",
    "Simulation",
    "im",
    "sm",
]



"""
Continuous-time models for induction machine drives.

Peak-valued complex space vectors are used. The space vector models are
implemented in stator coordinates. 

"""
import numpy as np
from motulator._helpers import abc2complex, complex2abc
from motulator._utils import Bunch


# %%
class InductionMachine:
    """
    Γ-equivalent model of an induction machine.

    An induction machine is modeled using the Γ-equivalent model [#Sle1989]_. 
    The model is implemented in stator coordinates. The flux linkages are used 
    as state variables.

    Parameters
    ----------
    n_p : int
        Number of pole pairs.
    R_s : float
        Stator resistance (Ω).
    R_r : float
        Rotor resistance (Ω).
    L_ell : float
        Leakage inductance (H).
    L_s : float
        Stator inductance (H).

    Notes
    -----
    The Γ model is chosen here since it can be extended with the magnetic
    saturation model in a staightforward manner. If the magnetic saturation is
    omitted, the Γ model is mathematically identical to the inverse-Γ and T
    models [#Sle1989]_.

    References
    ----------
    .. [#Sle1989] Slemon, "Modelling of induction machines for electric drives," 
       IEEE Trans. Ind. Appl., 1989, https://doi.org/10.1109/28.44251.

    """

    def __init__(self, n_p, R_s, R_r, L_ell, L_s):
        # pylint: disable=too-many-arguments
        self.n_p = n_p
        self.R_s, self.R_r = R_s, R_r
        self.L_ell, self.L_s = L_ell, L_s
        # Initial values
        self.psi_ss0, self.psi_rs0 = 0j, 0j

    def currents(self, psi_ss, psi_rs):
        """
        Compute the stator and rotor currents.

        Parameters
        ----------
        psi_ss : complex
            Stator flux linkage (Vs).
        psi_rs : complex
            Rotor flux linkage (Vs).

        Returns
        -------
        i_ss : complex
            Stator current (A).
        i_rs : complex
            Rotor current (A).

        """
        i_rs = (psi_rs - psi_ss)/self.L_ell
        i_ss = psi_ss/self.L_s - i_rs

        return i_ss, i_rs

    def magnetic(self, psi_ss, psi_rs):
        """
        Magnetic model.

        Parameters
        ----------
        psi_ss : complex
            Stator flux linkage (Vs).
        psi_rs : complex
            Rotor flux linkage (Vs).

        Returns
        -------
        i_ss : complex
            Stator current (A).
        i_rs : complex
            Rotor current (A).
        tau_M : float
            Electromagnetic torque (Nm).

        """
        i_ss, i_rs = self.currents(psi_ss, psi_rs)
        tau_M = 1.5*self.n_p*np.imag(i_ss*np.conj(psi_ss))

        return i_ss, i_rs, tau_M

    def f(self, psi_ss, psi_rs, u_ss, w_M):
        """
        Compute the state derivatives.

        Parameters
        ----------
        psi_ss : complex
            Stator flux linkage (Vs).
        psi_rs : complex
            Rotor flux linkage (Vs).
        u_ss : complex
            Stator voltage (V).
        w_M : float
            Rotor angular speed (mechanical rad/s).

        Returns
        -------
        complex list, length 2
            Time derivative of the state vector, [dpsi_ss, dpsi_rs]
        i_ss : complex
            Stator current (A).
        tau_M : float
            Electromagnetic torque (Nm).

        Notes
        -----
        In addition to the state derivatives, this method also returns the
        output signals (stator current `i_ss` and torque `tau_M`) needed for
        interconnection with other subsystems. This avoids overlapping
        computation in simulation.

        """
        i_ss, i_rs, tau_M = self.magnetic(psi_ss, psi_rs)
        dpsi_ss = u_ss - self.R_s*i_ss
        dpsi_rs = -self.R_r*i_rs + 1j*self.n_p*w_M*psi_rs

        return [dpsi_ss, dpsi_rs], i_ss, tau_M

    def meas_currents(self):
        """
        Measure the phase currents at the end of the sampling period.

        Returns
        -------
        i_s_abc : 3-tuple of floats
            Phase currents (A).

        """
        # Stator current space vector in stator coordinates
        i_ss, _ = self.currents(self.psi_ss0, self.psi_rs0)
        # Phase currents
        i_s_abc = complex2abc(i_ss)  # + noise + offset ...
        return i_s_abc


# %%
class InductionMachineSaturated(InductionMachine):
    """
    Γ-equivalent model of an induction machine model with main-flux saturation.

    This extends the InductionMachine class with a main-flux magnetic saturation
    model::

        L_s = L_s(abs(psi_ss))

    Parameters
    ----------
    n_p : int
        Number of pole pairs.
    R_s : float
        Stator resistance (Ω).
    R_r : float
        Rotor resistance (Ω).
    L_ell : float
        Leakage inductance (H).
    L_s : callable
        Stator inductance (H) as a function of the stator-flux magnitude.

    """

    def currents(self, psi_ss, psi_rs):
        """Override the base class method."""
        # Saturated value of the stator inductance.
        L_s = self.L_s(np.abs(psi_ss))
        # Currents
        i_rs = (psi_rs - psi_ss)/self.L_ell
        i_ss = psi_ss/L_s - i_rs
        return i_ss, i_rs


# %%
class InductionMachineInvGamma(InductionMachine):
    """
    Inverse-Γ model of an induction machine.

    This extends the InductionMachine class (based on the Γ model) by providing
    an interface for the inverse-Γ model parameters. Linear magnetics are 
    assumed. If magnetic saturation is to be modeled, the Γ model is preferred.

    Parameters
    ----------
    n_p : int
        Number of pole pairs.
    R_s : float
        Stator resistance (Ω).
    R_R : float
        Rotor resistance (Ω).
    L_sgm : float
        Leakage inductance (H).
    L_M : float
        Magnetizing inductance (H).

    """

    def __init__(self, n_p, R_s, R_R, L_sgm, L_M):
        # pylint: disable=too-many-arguments
        # Convert the inverse-Γ parameters to the Γ parameters
        gamma = L_M/(L_M + L_sgm)  # Magnetic coupling factor
        super().__init__(n_p, R_s, R_R/gamma**2, L_sgm/gamma, L_M + L_sgm)
        # Initial values
        self.psi_ss0 = 0j
        self.psi_rs0 = 0j  # self.psi_rs0 = self.psi_Rs0/gamma


# %%
class Drive:
    """
    Continuous-time model for an induction machine drive.

    This interconnects the subsystems of an induction machine drive and provides
    an interface to the solver. More complicated systems could be modeled using
    a similar template.

    Parameters
    ----------
    machine : InductionMachine | InductionMachineSaturated
        Induction machine model.
    mechanics : Mechanics
        Mechanics model.
    converver : Inverter
        Inverter model.

    """

    def __init__(self, machine=None, mechanics=None, converter=None):
        self.machine = machine
        self.mechanics = mechanics
        self.converter = converter
        self.t0 = 0  # Initial time
        # Store the solution in these lists
        self.data = Bunch()  # Stores the solution data
        self.data.t, self.data.q = [], []
        self.data.psi_ss, self.data.psi_rs = [], []
        self.data.theta_M, self.data.w_M = [], []

    def get_initial_values(self):
        """
        Get the initial values.

        Returns
        -------
        x0 : complex list, length 4
            Initial values of the state variables.

        """
        x0 = [
            self.machine.psi_ss0,
            self.machine.psi_rs0,
            self.mechanics.w_M0,
            self.mechanics.theta_M0,
        ]
        return x0

    def set_initial_values(self, t0, x0):
        """
        Set the initial values.

        Parameters
        ----------
        t0 : float
            Initial time (s).
        x0 : complex ndarray
            Initial values of the state variables.

        """
        self.t0 = t0
        self.machine.psi_ss0 = x0[0]
        self.machine.psi_rs0 = x0[1]
        # x0[2].imag and x0[3].imag are always zero
        self.mechanics.w_M0 = x0[2].real
        # Limit theta_M0 = x0[3].real into [-pi, pi)
        self.mechanics.theta_M0 = np.mod(x0[3].real + np.pi, 2*np.pi) - np.pi

    def f(self, t, x):
        """
        Compute the complete state derivative list for the solver.

        Parameters
        ----------
        t : float
            Time (s).
        x : complex ndarray
            State vector.

        Returns
        -------
        complex list
            State derivatives.

        """
        # Unpack the states
        psi_ss, psi_rs, w_M, _ = x
        # Interconnections: outputs for computing the state derivatives
        u_ss = self.converter.ac_voltage(
            self.converter.q, self.converter.u_dc0)
        # State derivatives plus the outputs for interconnections
        machine_f, _, tau_M = self.machine.f(psi_ss, psi_rs, u_ss, w_M)
        mechanics_f = self.mechanics.f(t, w_M, tau_M)
        # List of state derivatives
        return machine_f + mechanics_f

    def save(self, sol):
        """
        Save the solution.

        Parameters
        ----------
        sol : Bunch object
            Solution from the solver.

        """
        self.data.t.extend(sol.t)
        self.data.q.extend(sol.q)
        self.data.psi_ss.extend(sol.y[0])
        self.data.psi_rs.extend(sol.y[1])
        self.data.w_M.extend(sol.y[2].real)
        self.data.theta_M.extend(sol.y[3].real)

    def post_process(self):
        """Transform the lists to the ndarray format and post-process them."""
        # From lists to the ndarray
        for key in self.data:
            self.data[key] = np.asarray(self.data[key])

        # Some useful variables
        self.data.i_ss, _, self.data.tau_M = self.machine.magnetic(
            self.data.psi_ss, self.data.psi_rs)
        self.data.theta_M = np.mod(  # Limit into [-pi, pi)
            self.data.theta_M + np.pi, 2*np.pi) - np.pi
        self.data.theta_m = np.mod(  # Limit into [-pi, pi)
            self.machine.n_p*self.data.theta_M + np.pi, 2*np.pi) - np.pi
        self.data.w_m = self.machine.n_p*self.data.w_M
        self.data.tau_L = (
            self.mechanics.tau_L_t(self.data.t) +
            self.mechanics.tau_L_w(self.data.w_M))
        self.data.u_ss = self.converter.ac_voltage(
            self.data.q, self.converter.u_dc0)

        # Compute the inverse-Γ rotor flux
        try:
            # Saturable stator inductance
            L_s = self.machine.L_s(np.abs(self.data.psi_ss))
        except TypeError:
            # Constant stator inductance
            L_s = self.machine.L_s
        gamma = L_s/(L_s + self.machine.L_ell)  # Magnetic coupling factor
        self.data.psi_Rs = gamma*self.data.psi_rs


# %%
class DriveWithDiodeBridge(Drive):
    """
    Induction machine drive equipped with a diode bridge.

    This model extends the DriveWithDiodeBridge class with a model for a
    three-phase diode bridge fed from stiff supply voltages. The DC bus is
    modeled as an inductor and a capacitor.

    Parameters
    ----------
    machine : InductionMachine | InductionMachineSaturated
        Induction machine model.
    mechanics : Mechanics
        Mechanics model.
    converter : FrequencyConverter
        Frequency converter model.

    """

    def __init__(self, machine=None, mechanics=None, converter=None):
        super().__init__(machine, mechanics, converter)
        self.converter = converter
        self.data.u_dc, self.data.i_L = [], []

    def get_initial_values(self):
        """Extend the base class."""
        x0 = super().get_initial_values() + [
            self.converter.u_dc0, self.converter.i_L0
        ]
        return x0

    def set_initial_values(self, t0, x0):
        """Extend the base class."""
        super().set_initial_values(t0, x0[0:4])
        self.converter.u_dc0 = x0[4].real
        self.converter.i_L0 = x0[5].real

    def f(self, t, x):
        """Override the base class."""
        # Unpack the states for better readability
        psi_ss, psi_rs, w_M, _, u_dc, i_L = x

        # Interconnections: outputs for computing the state derivatives
        u_ss = self.converter.ac_voltage(self.converter.q, u_dc)
        mach_f, i_ss, tau_M = self.machine.f(psi_ss, psi_rs, u_ss, w_M)
        i_dc = self.converter.dc_current(self.converter.q, i_ss)

        # Return the list of state derivatives
        return (
            mach_f + self.mechanics.f(t, w_M, tau_M) +
            self.converter.f(t, u_dc, i_L, i_dc))

    def save(self, sol):
        """Extend the base class."""
        super().save(sol)
        self.data.u_dc.extend(sol.y[4].real)
        self.data.i_L.extend(sol.y[5].real)

    def post_process(self):
        """Extend the base class."""
        super().post_process()
        # From lists to the ndarray
        self.data.u_dc = np.asarray(self.data.u_dc)
        self.data.i_L = np.asarray(self.data.i_L)
        # Some useful variables
        self.data.u_ss = self.converter.ac_voltage(self.data.q, self.data.u_dc)
        self.data.i_dc = self.converter.dc_current(self.data.q, self.data.i_ss)
        u_g_abc = self.converter.grid_voltages(self.data.t)
        self.data.u_g = abc2complex(u_g_abc)
        # Voltage at the output of the diode bridge
        self.data.u_di = np.amax(u_g_abc, axis=0) - np.amin(u_g_abc, axis=0)
        # Diode briddge switching states (-1, 0, 1)
        q_g_abc = ((np.amax(u_g_abc, axis=0) == u_g_abc).astype(int) -
                   (np.amin(u_g_abc, axis=0) == u_g_abc).astype(int))
        # Grid current space vector
        self.data.i_g = abc2complex(q_g_abc)*self.data.i_L


# %%
class DriveTwoMassMechanics(Drive):
    """
    Induction machine drive with two-mass mechanics.

    This interconnects the subsystems of an induction machine drive and provides
    an interface to the solver.

    Parameters
    ----------
    machine : InductionMachine | InductionMachineSaturated
        Induction machine model.
    mechanics : MechanicsTwoMass
        Mechanics model.
    converter : Inverter
        Inverter model.

    """

    def __init__(self, machine=None, mechanics=None, converter=None):
        super().__init__(machine, mechanics, converter)
        self.data.w_L, self.data.theta_ML = [], []

    def get_initial_values(self):
        """Extend the base class."""
        x0 = super().get_initial_values() + [
            self.mechanics.w_L0, self.mechanics.theta_ML0
        ]
        return x0

    def set_initial_values(self, t0, x0):
        """Extend the base class."""
        super().set_initial_values(t0, x0[0:4])
        self.mechanics.w_L0 = x0[4].real
        self.mechanics.theta_ML0 = np.mod(x0[5].real + np.pi, 2*np.pi) - np.pi

    def f(self, t, x):
        """Override the base class."""
        # Unpack the states
        psi_ss, psi_rs, w_M, _, w_L, theta_ML = x
        # Interconnections: outputs for computing the state derivatives
        u_ss = self.converter.ac_voltage(
            self.converter.q, self.converter.u_dc0)
        # State derivatives plus the outputs for interconnections
        machine_f, _, tau_M = self.machine.f(psi_ss, psi_rs, u_ss, w_M)
        mechanics_f = self.mechanics.f(t, w_M, w_L, theta_ML, tau_M)
        # List of state derivatives
        return machine_f + mechanics_f

    def save(self, sol):
        """Extend the base class."""
        super().save(sol)
        self.data.w_L.extend(sol.y[4].real)
        self.data.theta_ML.extend(sol.y[5].real)

    def post_process(self):
        """Extend the base class."""
        super().post_process()
        # From lists to the ndarray
        self.data.w_L = np.asarray(self.data.w_L)
        self.data.theta_ML = np.asarray(self.data.theta_ML)



"""This package contains example controllers for synchronous machines."""

from motulator.control.sm._vector import (
    CurrentReferencePars,
    CurrentReference,
    CurrentCtrl,
    ModelPars,
    VectorCtrl,
)

from motulator.control.sm._observers import (Observer, FluxObserver)

from motulator.control.sm._flux_vector import (
    FluxVectorCtrl, FluxTorqueReference, FluxTorqueReferencePars)
from motulator.control.sm._obs_vhz import (
    ObserverBasedVHzCtrl, ObserverBasedVHzCtrlPars)
from motulator.control.sm._torque import TorqueCharacteristics
from motulator.control.sm._signal_inj import (
    SignalInjectionCtrl, SignalInjection)

__all__ = [
    "CurrentReferencePars", "CurrentReference", "CurrentCtrl", "ModelPars",
    "Observer", "VectorCtrl", "Observer", "FluxObserver", "FluxVectorCtrl",
    "FluxTorqueReference", "FluxTorqueReferencePars", "ObserverBasedVHzCtrl",
    "ObserverBasedVHzCtrlPars", "FluxObserver", "TorqueCharacteristics",
    "SignalInjectionCtrl", "SignalInjection"
]



"""
V/Hz control for induction motor drives.

The method is similar to [#Hin2022]_. Open-loop V/Hz control can be obtained as 
a special case by choosing::

    R_s, R_R = 0, 0
    k_u, k_w = 0, 0

References
----------
.. [#Hin2022] Hinkkanen, Tiitinen, Mölsä, Harnefors, "On the stability of
   volts-per-hertz control for induction motors," IEEE J. Emerg. Sel. Topics
   Power Electron., 2022, https://doi.org/10.1109/JESTPE.2021.3060583

"""
# %%
import numpy as np
from motulator.control._common import Ctrl, PWM
from motulator._helpers import abc2complex
from motulator._utils import Bunch


# %%
class VHzCtrl(Ctrl):
    """
    V/Hz control with the stator current feedback.

    Parameters
    ----------
    par : ModelPars
        Control parameters.

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, T_s, par, psi_s_nom, k_u=1., k_w=4., six_step=False):
        super().__init__()
        self.T_s = T_s
        self.w_m_ref = callable
        # Motor parameters
        self.R_s, self.R_R = par.R_s, par.R_R
        self.L_sgm, self.L_M = par.L_sgm, par.L_M
        # Control parameters
        self.k_u, self.k_w = k_w, k_u
        self.psi_s_ref = psi_s_nom
        w_rb = self.R_R*(self.L_M + self.L_sgm)/(self.L_sgm*self.L_M)
        self.alpha_f = .1*w_rb
        self.alpha_i = .1*w_rb
        # Instantiate classes
        self.pwm = PWM(six_step=six_step)
        self.rate_limiter = callable
        self.six_step = six_step
        # States
        self.i_s_ref, self.theta_s, self.w_r_ref = 0j, 0, 0
        # self.T_s0 = T_s
        # self.N = 100

    def __call__(self, mdl):
        """
        Run the main control loop.

        Parameters
        ----------
        mdl : Drive
            Continuous-time model for getting the feedback signals.

        Returns
        -------
        T_s : float
            Sampling period (s).
        d_abc : ndarray, shape (3,)
            Duty ratios.

        """
        # Rate limit the frequency reference
        w_m_ref = self.rate_limiter(self.T_s, self.w_m_ref(self.clock.t))

        # Measure the feedback signals
        i_s_abc = mdl.machine.meas_currents()  # Phase currents
        u_dc = mdl.converter.meas_dc_voltage()  # DC-bus voltage

        # Space vector transformation
        theta_s = self.theta_s
        i_s = np.exp(-1j*theta_s)*abc2complex(i_s_abc)

        # Slip compensation
        w_s_ref = w_m_ref + self.w_r_ref

        # Dynamic stator frequency and slip frequency
        w_s, w_r = self._stator_freq(w_s_ref, i_s)

        # Voltage reference
        u_s_ref = self._voltage_reference(w_s, i_s)

        # TODO: Synchronize the PWM to the stator frequency in overmodulation,
        # a simplistic approach given below.
        # f_s = w_s/(2*np.pi)  # Stator frequency
        # if self.six_step and self.N*np.abs(f_s) > 1/self.T_s0:
        #     self.T_s = 1/(self.N*f_s)
        # else:
        #     self.T_s = self.T_s0

        # Data logging
        data = Bunch(
            i_s=i_s,
            i_s_ref=self.i_s_ref,
            psi_s_ref=self.psi_s_ref,
            t=self.clock.t,
            theta_s=self.theta_s,
            u_dc=u_dc,
            u_s=self.pwm.realized_voltage,
            w_m_ref=w_m_ref,
            w_r=w_r,
            w_s=w_s,
        )
        self.save(data)

        # Update the states. Note that the low-pass-filtered values are marked
        # with ref at the end of the variable name.
        self.i_s_ref += self.T_s*self.alpha_i*(i_s - self.i_s_ref)
        self.w_r_ref += self.T_s*self.alpha_f*(w_r - self.w_r_ref)
        self.theta_s += self.T_s*w_s  # Next line: limit into [-pi, pi)
        self.theta_s = np.mod(self.theta_s + np.pi, 2*np.pi) - np.pi
        self.clock.update(self.T_s)

        # Compute the duty ratios
        d_abc = self.pwm(self.T_s, u_s_ref, u_dc, theta_s, w_s)

        return self.T_s, d_abc

    def _stator_freq(self, w_s_ref, i_s):
        # Compute the stator frequency for coordinate transformations

        # Operating-point quantities
        psi_R_ref = self.psi_s_ref - self.L_sgm*self.i_s_ref
        psi_R_ref_sqr = np.abs(psi_R_ref)**2
        # Compute the dynamic stator frequency
        if psi_R_ref_sqr > 0:
            # Slip estimate based on the measured current
            w_r = self.R_R*np.imag(i_s*np.conj(psi_R_ref))/psi_R_ref_sqr
            # Dynamic frequency
            w_s = w_s_ref + self.k_w*(self.w_r_ref - w_r)
        else:
            w_s, w_r = 0, 0
        return w_s, w_r

    def _voltage_reference(self, w_s, i_s):
        # Compute the stator voltage reference

        # Nominal magnetizing current
        i_sd_nom = self.psi_s_ref/(self.L_M + self.L_sgm)
        # Operating-point current for RI compensation
        i_s_ref0 = i_sd_nom + 1j*self.i_s_ref.imag
        # Term -R_s omitted to avoid problems due to the voltage saturation
        # k = -R_s + k_u*L_sgm*(alpha + 1j*w_m0)
        k = self.k_u*self.L_sgm*(self.R_R/self.L_M + 1j*w_s)
        u_s_ref = (
            self.R_s*i_s_ref0 + 1j*w_s*self.psi_s_ref + k*(self.i_s_ref - i_s))
        return u_s_ref



"""This package contains example controllers."""

from motulator.control._common import (
    ComplexPICtrl, RateLimiter, SpeedCtrl, PICtrl, PWM)

import motulator.control.sm as sm
import motulator.control.im as im

__all__ = [
    "ComplexPICtrl", "RateLimiter", "SpeedCtrl", "PICtrl", "PWM", "sm", "im"
]



"""
State observers for synchronous machines.
"""
import numpy as np


# %%
class Observer:
    """
    Observer for synchronous machines in estimated rotor coordinates.

    This observer estimates the rotor angle, the rotor speed, and the stator 
    flux linkage. The design is based on [#Hin2018]_. The observer gain 
    decouples the electrical and mechanical dynamics and allows placing the 
    poles of the corresponding linearized estimation error dynamics. This 
    implementation operates in estimated rotor coordinates. The observer can 
    also be used in the sensored mode by providing the measured rotor speed as 
    an input.

    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    alpha_o : float, optional
        Observer bandwidth (electrical rad/s). The default is 2*pi*40.
    k : callable, optional
        Observer gain as a function of the rotor angular speed. The default is 
        ``lambda w_m: 0.25*(R_s*(L_d + L_q)/(L_d*L_q) + 0.2*abs(w_m))`` if
        `sensorless` else ``lambda w_m: 2*pi*15``.

    Attributes
    ----------
    theta_m : float
        Rotor angle estimate (electrical rad).
    w_m : float
        Rotor speed estimate (electrical rad/s).
    psi_s : complex
        Stator flux estimate (Vs).

    References
    ----------
    .. [#Hin2018] Hinkkanen, Saarakkala, Awan, Mölsä, Tuovinen, "Observers for
       sensorless synchronous motor drives: Framework for design and analysis,"
       IEEE Trans. Ind. Appl., 2018, https://doi.org/10.1109/TIA.2018.2858753

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, par, alpha_o=2*np.pi*40, k=None, sensorless=True):
        self.R_s = par.R_s
        self.L_d, self.L_q, self.psi_f = par.L_d, par.L_q, par.psi_f
        self.psi_f = par.psi_f
        self.alpha_o = alpha_o
        self.sensorless = sensorless
        self.k1 = k
        if self.sensorless:
            if self.k1 is None:  # If not given, use the default gains
                sigma0 = .25*par.R_s*(par.L_d + par.L_q)/(par.L_d*par.L_q)
                self.k1 = lambda w_m: (sigma0 + .2*np.abs(w_m))
            self.k2 = self.k1
        else:
            if self.k1 is None:
                self.k1 = lambda w_m: 2*np.pi*15
            self.k2 = 0*self.k1
        # Initial states
        self.theta_m, self.w_m, self.psi_s = 0, 0, par.psi_f

    def update(self, T_s, u_s, i_s, w_m=None):
        """
        Update the states for the next sampling period.

        Parameters
        ----------
        T_s : float
            Sampling period (s).
        u_s : complex
            Stator voltage (V) in estimated rotor coordinates.
        i_s : complex
            Stator current (A) in estimated rotor coordinates.
        w_m : float, optional
            Rotor angular speed (electrical rad/s). Needed only in the sensored
            mode. The default is None. 

        """

        # Estimation error
        e = self.L_d*i_s.real + 1j*self.L_q*i_s.imag + self.psi_f - self.psi_s

        if self.sensorless:
            # Auxiliary flux
            psi_a = self.psi_f + (self.L_d - self.L_q)*np.conj(i_s)

            # Observer gains
            k1 = self.k1(self.w_m)
            k2 = k1*psi_a/np.conj(psi_a) if np.abs(psi_a) > 0 else k1

            # Speed estimation
            eps = -np.imag(e/psi_a) if np.abs(psi_a) > 0 else 0
            w_s = 2*self.alpha_o*eps + self.w_m
        else:
            k1, k2 = self.k1, 0
            w_s = w_m
            eps = 0

        # Update the states
        self.psi_s += T_s*(
            u_s - self.R_s*i_s - 1j*w_s*self.psi_s + k1*e + k2*np.conj(e))
        self.w_m += T_s*self.alpha_o**2*eps
        self.theta_m += T_s*w_s  # Next line: limit into [-pi, pi)
        self.theta_m = np.mod(self.theta_m + np.pi, 2*np.pi) - np.pi


# %%
class FluxObserver:
    """
    Sensorless stator flux observer in external coordinates.

    This observer estimates the stator flux linkage and the angle of the coordinate
    system with respect to the d-axis of the rotor. Speed-estimation is omitted. 
    The observer gain decouples the electrical and mechanical dynamics and allows placing 
    the poles of the corresponding linearized estimation error dynamics. This implementation operates 
    in external coordinates (typically synchronous coordinates defined by reference signals 
    of a control system).

    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    alpha_o : float, optional
        Observer gain (rad/s). The default is 2*pi*20.
    zeta_inf : float, optional
        Damping ratio at infinite speed. The default is 0.2.

    Attributes
    ----------
    delta : float
        Angle estimate of the coordinate system with respect
        to the d-axis of the rotor (electrical rad).
    psi_s : complex
        Stator flux estimate (Vs).
  
    References
    ----------
    .. [#Tii2022] Tiitinen, Hinkkanen, Kukkola, Routimo, Pellegrino, Harnefors, 
       "Stable and passive observer-based V/Hz control for synchronous Motors," 
       Proc. IEEE ECCE, Detroit, MI, Oct. 2022, 
       https://doi.org/10.1109/ECCE50734.2022.9947858

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, par, alpha_o=2*np.pi*20, zeta_inf=.2):
        self.R_s = par.R_s
        self.L_d = par.L_d
        self.L_q = par.L_q
        self.psi_f = par.psi_f
        self.alpha_o = alpha_o
        self.b_p = .5*par.R_s*(par.L_d + par.L_q)/(par.L_d*par.L_q)
        self.zeta_inf = zeta_inf
        # Initial states
        self.delta, self.psi_s = 0, par.psi_f

    def update(self, T_s, u_s, i_s, w_s):
        """
        Update the states for the next sampling period.

        Parameters
        ----------
        T_s : float
            Sampling period (s).
        u_s : complex
            Stator voltage (V).
        i_s : complex
            Stator current (A).
        w_s : float
            Stator angular frequency (rad/s).

        """
        # Transformations to rotor coordinates. This is mathematically
        # equivalent to the version in [Tii2022].
        i_sr = i_s*np.exp(1j*self.delta)
        psi_sr = self.psi_s*np.exp(1j*self.delta)

        # Auxiliary flux and estimation error in rotor coordinates
        psi_ar = self.psi_f + (self.L_d - self.L_q)*np.conj(i_sr)
        e_r = self.L_d*i_sr.real + 1j*self.L_q*i_sr.imag + self.psi_f - psi_sr

        # Auxiliary flux in controller coordinates
        psi_a = np.exp(-1j*self.delta)*psi_ar

        k = self.b_p + 2*self.zeta_inf*np.abs(w_s)

        if np.abs(psi_ar) > 0:
            # Correction voltage in controller coordinates
            v = k*psi_a*np.real(e_r/psi_ar)
            # Error signal
            w_delta = self.alpha_o*np.imag(e_r/psi_ar)
        else:
            v, w_delta = 0, 0

        # Update the states
        self.psi_s += T_s*(u_s - self.R_s*i_s - 1j*w_s*self.psi_s + v)
        self.delta += T_s*w_delta



"""Continuous-time synchronous machine models."""

from motulator.model.sm._drive import (
    Drive,
    DriveTwoMassMechanics,
    SynchronousMachine,
    SynchronousMachineSaturated,
)

from motulator.model.sm._flux_maps import (
    import_syre_data,
    plot_flux_map,
    plot_flux_vs_current,
)

__all__ = [
    "Drive",
    "DriveTwoMassMechanics",
    "SynchronousMachine",
    "SynchronousMachineSaturated",
    "import_syre_data",
    "plot_flux_map",
    "plot_flux_vs_current",
]



"""
Observer-based V/Hz control for induction machine drives.

This implements the observer-based V/Hz control method described in [#Tii2022]_. 
The state-feedback control law is in the alternative form which uses an
intermediate stator current reference.

References
----------
.. [#Tii2022] Tiitinen, Hinkkanen, Harnefors, "Stable and passive observer-based 
   V/Hz control for induction motors," Proc. IEEE ECCE, Detroit, MI, Oct. 2022,
   https://doi.org/10.1109/ECCE50734.2022.9948057

"""

# %%
from dataclasses import dataclass
import numpy as np
from motulator.control._common import Ctrl, PWM, RateLimiter
from motulator.control.im._observers import FluxObserver
from motulator._helpers import abc2complex
from motulator._utils import Bunch


# %%
# pylint: disable=too-many-instance-attributes
@dataclass
class ObserverBasedVHzCtrlPars:
    """
    Parameters for the control system.

    Parameters
    ----------
    psi_s_nom : float
        Nominal stator flux linkage (Vs). 
    i_s_max : float, optional
        Maximum stator current (A). The default is inf.
    k_tau : float, optional
        Torque controller gain. The default is 3.
    alpha_psi : float, optional
        Stator flux control bandwidth (rad/s). The default is 2*pi*20.
    alpha_f : float, optional
        Torque high-pass filter bandwidth (rad/s). The default is 2*pi*1.
    alpha_r : float, optional
        Low-pass-filter bandwidth (rad/s) for slip angular frequency. The
        default is 2*pi*1.
    slip_compensation : bool, optional
        Enable slip compensation. The default is False.

    """
    # par: InitVar[ModelPars | None] = None
    psi_s_nom: float = None
    i_s_max: float = np.inf
    k_tau: float = 3.
    alpha_psi: float = 2*np.pi*20
    alpha_f: float = 2*np.pi*1
    alpha_r: float = 2*np.pi*1
    slip_compensation: bool = False


# %%
class ObserverBasedVHzCtrl(Ctrl):
    """
    Observer-based V/Hz control for induction machines.

    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    ctrl_par : ObserverBasedVHzCtrlPars
        Control system parameters.
    T_s : float, optional
        Sampling period (s). The default is 250e-6.

    Attributes
    ----------
    observer : SensorlessObserverExtCoord
        Sensorless reduced-order flux observer in external coordinates.
    rate_limiter : RateLimiter
        Rate limiter for the speed reference.
    pwm : PWM
        Pulse-width modulation.
    w_m_ref : callable
        Speed reference (electrical rad/s) as a function of time (s).

    """

    # pylint: disable=too-many-instance-attributes, too-many-arguments
    def __init__(self, par, ctrl_par, T_s=250e-6):
        super().__init__()
        self.T_s = T_s
        # Model parameters
        self.R_s, self.R_R = par.R_s, par.R_R
        self.L_sgm = par.L_sgm
        self.n_p = par.n_p
        # References
        self.w_m_ref = callable
        self.psi_s_ref = ctrl_par.psi_s_nom
        self.i_s_max = ctrl_par.i_s_max
        # Control parameters
        self.slip_compensation = ctrl_par.slip_compensation
        self.alpha_f = ctrl_par.alpha_f
        self.alpha_r = ctrl_par.alpha_r
        self.alpha_psi = ctrl_par.alpha_psi
        self.k_tau = ctrl_par.k_tau
        # Instantiate classes
        self.observer = FluxObserver(par, alpha_o=2*np.pi*40)
        self.rate_limiter = RateLimiter()
        self.pwm = PWM(six_step=False)
        # States
        self.theta_s, self.tau_M_ref, self.w_r_ref = 0, 0, 0

    def __call__(self, mdl):
        """
        Run the main control loop.

        Parameters
        ----------
        mdl : Drive
            Continuous-time system model for getting the feedback signals.

        Returns
        -------
        T_s : float
            Sampling period (s).
        d_abc : ndarray, shape (3,)
            Duty ratios.

        """
        # Get the speed reference
        w_m_ref = self.rate_limiter(self.T_s, self.w_m_ref(self.clock.t))

        # Measure the feedback signals
        i_s_abc = mdl.machine.meas_currents()  # Phase currents
        u_dc = mdl.converter.meas_dc_voltage()  # DC-bus voltage

        # Get the states
        u_s = self.pwm.realized_voltage
        psi_R = self.observer.psi_R
        tau_M_ref = self.tau_M_ref
        w_r_ref = self.w_r_ref
        theta_s = self.theta_s

        # Space vector and coordinate transformation
        i_s = np.exp(-1j*theta_s)*abc2complex(i_s_abc)

        # Torque estimate
        tau_M = 1.5*self.n_p*np.imag(i_s*np.conj(psi_R))

        # Slip frequency compensation (if enabled) for the low-pass filter.
        # Note, could also be based on the low-pass filtered torque.
        psi_R_sqr = np.abs(psi_R)**2
        if self.slip_compensation and psi_R_sqr > 0:
            w_r = self.R_R*tau_M/(1.5*self.n_p*psi_R_sqr)
        else:
            w_r = 0

        # Slip compensation. Uses the low-pass filtered slip estimate w_r_ref.
        # Note if slip compensation disabled w_r_ref == 0.
        w_s_ref = w_m_ref + w_r_ref

        # Dynamic frequency
        w_s = w_s_ref - self.k_tau*(tau_M - tau_M_ref)

        # State feedback
        u_s_ref = self._state_feedback(i_s, psi_R, w_s)

        # Data logging
        data = Bunch(
            i_s=i_s,
            psi_s=psi_R + self.L_sgm*i_s,
            psi_s_ref=self.psi_s_ref,
            t=self.clock.t,
            theta_s=self.theta_s,
            u_dc=u_dc,
            u_s=u_s,
            w_m=self.observer.w_m,
            w_m_ref=w_m_ref,
            w_s=w_s,
            tau_M=tau_M,
        )
        self.save(data)

        # Update the states
        self.observer.update(self.T_s, u_s, i_s, w_s)
        self.w_r_ref += self.T_s*self.alpha_r*(w_r - self.w_r_ref)
        self.tau_M_ref += self.T_s*self.alpha_f*(tau_M - self.tau_M_ref)
        self.theta_s += self.T_s*w_s  # Next line: limit into [-pi, pi)
        self.theta_s = np.mod(self.theta_s + np.pi, 2*np.pi) - np.pi
        self.clock.update(self.T_s)

        # Calculate the duty ratios and update the voltage estimate state
        d_abc = self.pwm(self.T_s, u_s_ref, u_dc, theta_s, w_s)

        return self.T_s, d_abc

    def _state_feedback(self, i_s, psi_R, w_s):
        # Internal current reference for state feedback
        i_s_ref = (self.psi_s_ref - psi_R)/self.L_sgm
        # Limit the reference
        if np.abs(i_s_ref) > self.i_s_max:
            i_s_ref = self.i_s_max*i_s_ref/np.abs(i_s_ref)
        # State feedback
        u_s_ref = (
            self.R_s*i_s_ref + 1j*w_s*self.psi_s_ref +
            self.L_sgm*self.alpha_psi*(i_s_ref - i_s))
        return u_s_ref



"""
Vector control methods for induction machine drives.

The algorithms are written based on the inverse-Γ model.

"""
from dataclasses import dataclass, InitVar
import numpy as np
from motulator._helpers import abc2complex
from motulator.control._common import PWM, Ctrl, ComplexPICtrl, SpeedCtrl
from motulator.control.im._observers import Observer, FullOrderObserver
from motulator._utils import Bunch


# %%
@dataclass
class ModelPars:
    """
    Inverse-Γ model parameters of an induction machine.
    
    Parameters
    ----------
    R_s : float
        Stator resistance (Ω).
    R_R : float
        Rotor resistance (Ω).
    L_sgm : float
        Leakage inductance (H).
    L_M : float
        Magnetizing inductance (H).
    n_p : int
        Number of pole pairs.  
    J : float
        Moment of inertia (kgm²).  
    
    """
    R_s: float = None
    R_R: float = None
    L_sgm: float = None
    L_M: float = None
    n_p: int = None
    J: float = None


# %%
# pylint: disable=too-many-instance-attributes
class VectorCtrl(Ctrl):
    """
    Vector control for induction machine drives.

    This class interconnects the subsystems of the control system and provides 
    the interface to the solver.

    Parameters
    ----------
    par : InductionMachinePars
        Machine model parameters.
    ref : CurrentReferencePars
        Parameters for reference generation.
    T_s : float, optional
        Sampling period (s). The default is 250e-6.
    sensorless : bool, optional
        If True, sensorless control is used. The default is True.

    Attributes
    ----------
    current_ref : CurrentReference
        Current reference generator.
    observer : Observer
        Flux and speed observer. 
    current_ctrl : CurrentCtrl
        Current controller.
    speed_ctrl : SpeedCtrl 
        Speed controller.
    pwm : PWM
        Pulse-width modulation.
    w_m_ref : callable
        Speed reference (electrical rad/s) as a function of time (s).

    """

    def __init__(self, par, ref, T_s=250e-6, sensorless=True):
        super().__init__()
        self.w_m_ref = callable
        self.T_s = T_s
        self.sensorless = sensorless
        self.n_p = par.n_p
        self.current_ref = CurrentReference(par, ref)
        self.current_ctrl = CurrentCtrl(par, 2*np.pi*200)
        self.observer = Observer(par, sensorless=sensorless)
        self.speed_ctrl = SpeedCtrl(par.J, 2*np.pi*4)
        self.pwm = PWM()

    # pylint: disable=too-many-locals
    def __call__(self, mdl):
        """
        Run the main control loop.

        Parameters
        ----------
        mdl : Drive
            Continuous-time system model for getting the feedback signals.

        Returns
        -------
        T_s : float
            Sampling period (s).
        d_abc : ndarray, shape (3,)
            Duty ratios.

        """
        # Get the speed reference
        w_m_ref = self.w_m_ref(self.clock.t)

        # Feedback signals
        i_s_abc = mdl.machine.meas_currents()
        u_dc = mdl.converter.meas_dc_voltage()

        # Get the states
        if self.sensorless:
            w_m = self.observer.w_m
        else:
            w_m = self.n_p*mdl.mechanics.meas_speed()
        u_s = self.pwm.realized_voltage
        psi_R, theta_s = self.observer.psi_R, self.observer.theta_s

        # Space vector and coordinate transformation
        i_s = np.exp(-1j*theta_s)*abc2complex(i_s_abc)

        # Controller outputs
        tau_M_ref = self.speed_ctrl.output(w_m_ref/self.n_p, w_m/self.n_p)
        i_s_ref, tau_M_ref_lim = self.current_ref.output(tau_M_ref, psi_R)
        u_s_ref = self.current_ctrl.output(i_s_ref, i_s)

        # Save data
        data = Bunch(
            i_s=i_s,
            i_s_ref=i_s_ref,
            psi_R=psi_R,
            t=self.clock.t,
            tau_M_ref_lim=tau_M_ref_lim,
            theta_s=theta_s,
            u_dc=u_dc,
            u_s=u_s,
            w_m=w_m,
            w_m_ref=w_m_ref,
        )
        self.save(data)

        # Compute the flux angular frequency and update the observer states
        # (w_m needed only in sensored mode)
        w_s = self.observer(self.T_s, u_s, i_s, w_m)

        # Update the controller states
        self.speed_ctrl.update(self.T_s, tau_M_ref_lim)
        self.current_ref.update(self.T_s, u_s_ref, u_dc)
        self.current_ctrl.update(self.T_s, u_s, w_s)
        self.clock.update(self.T_s)

        # Calculate the duty ratios and update the voltage estimate state
        d_abc = self.pwm(self.T_s, u_s_ref, u_dc, theta_s, w_s)

        return self.T_s, d_abc


# %%
class CurrentCtrl(ComplexPICtrl):
    """
    2DOF PI current controller for induction machines.

    This class provides an interface for a current controller for induction 
    machines. The gains are initialized based on the desired closed-loop 
    bandwidth and the leakage inductance. 

    Parameters
    ----------
    par : ModelPars
        Machine parameters, contains the leakage inductance `L_sgm` (H).  
    alpha_c : float
        Closed-loop bandwidth (rad/s).

    """

    def __init__(self, par, alpha_c):
        k_t = alpha_c*par.L_sgm
        k_i = alpha_c*k_t
        k_p = 2*k_t
        super().__init__(k_p, k_i, k_t)


# %%
# pylint: disable=too-many-instance-attributes
@dataclass
class CurrentReferencePars:
    """
    Parameters for reference generation.

    This dataclass stores the nominal and limit values needed for reference 
    generation. For calculating the rotor flux reference, the machine parameters 
    are also required.

    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    i_s_max : float
        Maximum stator current (A). 
    u_s_nom : float, optional
        Nominal stator voltage (V). The default is sqrt(2/3)*400.
    w_s_nom : float, optional
        Nominal stator angular frequency (rad/s). The default is 2*pi*50.
    psi_R_nom : float, optional
        Nominal rotor flux linkage (Vs). The default is 
        `(u_s_nom/w_s_nom)/(1 + L_sgm/L_M)`.
    k_fw : float, optional
        Field-weakening gain (1/H). The default is `2*R_R/(w_s_nom*L_sgm**2)`.
    k_u : float, optional
        Voltage utilization factor. The default is 0.95.
    
    """
    par: InitVar[ModelPars] = None
    i_s_max: float = None
    u_s_nom: InitVar[float] = np.sqrt(2/3)*400
    w_s_nom: InitVar[float] = 2*np.pi*50
    psi_R_nom: float = None
    k_fw: float = None
    k_u: float = 0.95

    def __post_init__(self, par, u_s_nom, w_s_nom):
        psi_s_nom = u_s_nom/w_s_nom  # Nominal stator flux
        if self.psi_R_nom is None:
            # Nominal rotor flux (omitting the slip)
            self.psi_R_nom = psi_s_nom/(1 + par.L_sgm/par.L_M)
        if self.k_fw is None:
            self.k_fw = 2*par.R_R/(w_s_nom*par.L_sgm**2)


# %%
class CurrentReference:
    """
    Current reference generation.

    In the base-speed region, the current reference in rotor-flux coordinates is
    given by::

        i_s_ref = psi_R_nom/L_M + 1j*tau_M_ref/(1.5*n_p*abs(psi_R))

    where `psi_R_nom` is the nominal rotor flux magnitude and `psi_R` is the 
    estimated rotor flux. The field-weakening operation is based on adjusting 
    the flux-producing current component::

        i_s_ref.real = (k_fw/s)*(u_s_max - abs(u_s_ref))

    where `1/s` refers to integration, ``u_s_max = k_u*u_dc/sqrt(3)`` is the 
    maximum stator voltage in the linear modulation region, `u_s_ref` is the 
    (unlimited) stator voltage reference, and `k_fw` is the field-weakening 
    gain. The field-weakening method and its tuning corresponds roughly to
    [#Hin2006]_. Furthermore, the torque-producing current component 
    `i_s_ref.imag` is limited based on the maximum stator current and the 
    breakdown slip. 

    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    ref : CurrentReferencePars
        Reference generation parameters.

    References
    ----------
    .. [#Hin2006] Hinkkanen, Luomi, "Braking scheme for vector-controlled 
       induction motor drives equipped with diode rectifier without braking 
       resistor," IEEE Trans. Ind. Appl., 2006, 
       https://doi.org/10.1109/TIA.2006.880852

    """

    def __init__(self, par, ref):
        # Machine model parameters
        self.L_sgm = par.L_sgm
        self.n_p = par.n_p
        # Other parameters
        self.i_s_max = ref.i_s_max
        self.k_u = ref.k_u
        # Field-weakening gain
        self.k_fw = ref.k_fw
        # Nominal d-axis current
        self.i_sd_nom = ref.psi_R_nom/par.L_M
        # State variable
        self.i_sd_ref = self.i_sd_nom

    def output(self, tau_M_ref, psi_R):
        """
        Compute the stator current reference.

        Parameters
        ----------
        tau_M_ref : float
            Torque reference (Nm).
        psi_R : float
            Rotor flux magnitude (Vs).

        Returns
        -------
        i_s_ref : complex
            Stator current reference (A).
        tau_M_ref_lim : float
            Limited torque reference (Nm).

        """

        def q_axis_current_limit(i_sd_ref, psi_R):
            # Priority given to the d component
            i_sq_max1 = np.sqrt(self.i_s_max**2 - i_sd_ref**2)
            # Breakdown torque limit
            i_sq_max2 = psi_R/self.L_sgm + i_sd_ref
            # q-axis current limit
            i_sq_max = np.min([i_sq_max1, i_sq_max2])
            return i_sq_max

        # q-axis current reference
        i_sq_ref = tau_M_ref/(1.5*self.n_p*psi_R) if psi_R > 0 else 0

        # Limit the current
        i_sq_max = q_axis_current_limit(self.i_sd_ref, psi_R)
        i_sq_ref = np.clip(i_sq_ref, -i_sq_max, i_sq_max)

        # Current reference
        i_s_ref = self.i_sd_ref + 1j*i_sq_ref

        # Limited torque (for the speed controller)
        tau_M_ref_lim = 1.5*self.n_p*psi_R*i_sq_ref

        return i_s_ref, tau_M_ref_lim

    def update(self, T_s, u_s_ref, u_dc):
        """
        Field-weakening based on the unlimited reference voltage.

        Parameters
        ----------
        T_s : float
            Sampling period (s)
        u_s_ref : complex
            Unlimited stator voltage reference (V).
        u_dc : float
            DC-bus voltage (V).

        """
        u_s_max = self.k_u*u_dc/np.sqrt(3)
        self.i_sd_ref += T_s*self.k_fw*(u_s_max - np.abs(u_s_ref))
        self.i_sd_ref = np.clip(self.i_sd_ref, -self.i_s_max, self.i_sd_nom)



"""Sensorless control with signal injection for synchronous machine drives."""

import numpy as np
from motulator._helpers import abc2complex
from motulator.control._common import Ctrl, PWM, SpeedCtrl
from motulator.control.sm._vector import CurrentCtrl, CurrentReference
from motulator._utils import Bunch


# %%
class SignalInjectionCtrl(Ctrl):
    """
    Sensorless control with signal injection for synchronous machine drives.

    This class implements a square-wave signal injection for low-speed 
    operation according to [#Kim2012]_. A phase-locked loop is used to track 
    the rotor position. 
    
    Notes
    -----
    For a wider speed range, signal injection could be combined to a 
    model-based observer. The effects of magnetic saturation are not
    compensated for in this version.

    References
    ----------
    .. [#Kim2012] Kim, Ha, Sul, "PWM switching frequency signal injection 
       sensorless method in IPMSM," IEEE Trans. Ind. Appl., 2012,
       https://doi.org/10.1109/TIA.2012.2210175

    Parameters
    ----------
    T_s : float
        Sampling period (s).
    pars : ModelPars
        Machine model parameters.
    U_inj : float
        Amplitude of the injected voltage (V).
    w_o : float
        PLL natural frequency (rad/s).

    Attributes
    ----------
    current_ctrl : CurrentCtrl
        Current controller.
    speed_ctrl : SpeedCtrl
        Speed controller.
    current_ref : CurrentReference
        Current reference generator.
    pll : PhaseLockedLoop
        Phase-locked loop.
    signal_inj : SignalInjection
        Signal injection.
    w_m_ref : callable
        Speed reference (electrical rad/s).
    pwm : PWM
        Pulse-width modulation.

    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, par, ref, T_s=250e-6):
        super().__init__()
        self.T_s = T_s
        self.n_p = par.n_p
        self.current_ref = CurrentReference(par, ref)
        self.current_ctrl = CurrentCtrl(par, 2*np.pi*200)
        self.pll = PhaseLockedLoop(w_o=2*np.pi*40)
        self.signal_inj = SignalInjection(par, U_inj=250)
        self.speed_ctrl = SpeedCtrl(par.J, 2*np.pi*4)
        self.pwm = PWM()
        self.w_m_ref = callable

    def __call__(self, mdl):
        """
        Run the main control loop.

        Parameters
        ----------
        mdl : Drive
            Continuous-time system model for getting the feedback signals.

        Returns
        -------
        T_s : float
            Sampling period (s).
        d_abc : ndarray, shape (3,)
            Duty ratios.

        """
        # Get the speed reference
        w_m_ref = self.w_m_ref(self.clock.t)

        # Measure the feedback signals
        i_s_abc = mdl.machine.meas_currents()  # Phase currents
        u_dc = mdl.converter.meas_dc_voltage()  # DC-bus voltage
        u_s = self.pwm.realized_voltage

        # Get the rotor speed and position estimates
        w_m, theta_m = self.pll.w_m, self.pll.theta_m

        # Current vector in estimated rotor coordinates
        i_s = np.exp(-1j*theta_m)*abc2complex(i_s_abc)

        # Filter the current measurement for the current controller
        i_s_filt = self.signal_inj.filter_current(i_s)

        # Outputs
        tau_M_ref = self.speed_ctrl.output(w_m_ref/self.n_p, w_m/self.n_p)
        i_s_ref, tau_M_ref_lim = self.current_ref.output(tau_M_ref, w_m, u_dc)
        err = self.signal_inj.output(self.T_s, i_s.imag)
        # Superimpose the excitation voltage on the d-axis
        u_s_ref = self.current_ctrl.output(
            i_s_ref, i_s_filt) + self.signal_inj.u_sd_inj

        # Data logging
        data = Bunch(
            i_s=i_s_filt,
            i_s_ref=i_s_ref,
            t=self.clock.t,
            tau_M_ref_lim=tau_M_ref_lim,
            theta_m=theta_m,
            u_dc=u_dc,
            u_s=u_s,
            w_m=w_m,
            w_m_ref=w_m_ref,
        )
        self.save(data)

        # Update states
        self.speed_ctrl.update(self.T_s, tau_M_ref_lim)
        self.current_ref.update(self.T_s, tau_M_ref_lim, u_s_ref, u_dc)
        self.current_ctrl.update(self.T_s, u_s, w_m)
        self.signal_inj.update(i_s)
        self.pll.update(self.T_s, err)
        self.clock.update(self.T_s)

        # PWM output
        d_abc = self.pwm(self.T_s, u_s_ref, u_dc, theta_m, w_m)

        return self.T_s, d_abc


# %%
class SignalInjection:
    """
    Estimate the rotor position error based on signal injection.

    This signal injection method estimates the rotor position error based on
    the injected switching frequency signal. The estimate can be used in a 
    phase-locked loop or in a state observer to robustify low-speed sensorless 
    operation.

    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    U_inj : float
        Injected voltage amplitude (V).

    """

    def __init__(self, par, U_inj):
        # Error gain
        self.k = .5*par.L_d*par.L_q/((par.L_q - par.L_d))
        # States
        self._i_s_old, self._i_s_older = 0, 0
        self.u_sd_inj = U_inj

    def output(self, T_s, i_sq):
        """
        Compute the rotor position estimation error.

        Parameters
        ----------
        T_s : float
            Sampling period (s).
        i_sq : float
            q-axis stator current (A) in estimated rotor coordinates.

        Returns
        -------
        err : float
            Rotor position estimation error (electrical rad).

        """
        di_sq = i_sq - 2*self._i_s_old.imag + self._i_s_older.imag
        err = (self.k/T_s)*di_sq/self.u_sd_inj if np.abs(
            self.u_sd_inj) > 0 else 0
        return err

    def update(self, i_s):
        """
        Store the old current values for the next sampling period.

        Parameters
        ----------
        i_s : complex
            Stator current in estimated rotor coordinates.

        """
        # Update the integral states
        self._i_s_older = self._i_s_old
        self._i_s_old = i_s
        # Reverse the d-axis square wave injection voltage
        self.u_sd_inj = -self.u_sd_inj

    def filter_current(self, i_s):
        """
        Filter the stator current using the previously measured value.

        Parameters
        ----------
        i_s : complex
            Unfiltered stator current (A) in estimated rotor coordinates.

        Returns
        -------
        i_s_filt : complex
            Filtered stator current (A) in estimated rotor coordinates.

        """
        # Filter currents
        i_s_filt = .5*(i_s + self._i_s_old)
        return i_s_filt


# %%
class PhaseLockedLoop:
    """
    Simple phase-locked loop for rotor-position estimation.

    Parameters
    ----------
    w_o : float
        Natural frequency (rad/s).

    """

    def __init__(self, w_o):
        # Gains
        self.k_p = w_o
        self.k_i = w_o**2
        # States
        self.theta_m, self.w_m = 0, 0

    def update(self, T_s, err):
        """
        Update the states for the next sampling period.

        Parameters
        ----------
        T_s : float
            Sampling period (s).
        err : float
            Rotor position error (rad).

        """
        # Speed estimation
        w_m = self.k_p*err + self.w_m

        # Update the states
        self.w_m += T_s*self.k_i*err
        self.theta_m += T_s*w_m  # Next line: limit into [-pi, pi)
        self.theta_m = np.mod(self.theta_m + np.pi, 2*np.pi) - np.pi



"""Utils."""


# %%
# The Bunch class is from sklearn.utils:
#
#    https://github.com/scikit-learn/scikit-learn
#
# Its license is provided below.
#
# ----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2007-2021 The scikit-learn developers.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
class Bunch(dict):
    """
    Container object exposing keys as attributes.

    Bunch objects are sometimes used as an output for functions and methods.
    They extend dictionaries by enabling values to be accessed by key,
    `bunch["value_key"]`, or by an attribute, `bunch.value_key`.

    Examples
    --------
    >>> from sklearn.utils import Bunch
    >>> b = Bunch(a=1, b=2)
    >>> b['b']
    2
    >>> b.b
    2
    >>> b.a = 3
    >>> b['a']
    3
    >>> b.c = 6
    >>> b['c']
    6

    """

    def __init__(self, **kwargs):
        super().__init__(kwargs)

    def __setattr__(self, key, value):
        self[key] = value

    def __dir__(self):
        return self.keys()

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            # pylint: disable=raise-missing-from
            raise AttributeError(key)

    def __setstate__(self, state):
        # Bunch pickles generated with scikit-learn 0.16.* have an non
        # empty __dict__. This causes a surprising behaviour when
        # loading these pickles scikit-learn 0.17: reading bunch.key
        # uses __dict__ but assigning to bunch.key use __setattr__ and
        # only changes bunch['key']. More details can be found at:
        # https://github.com/scikit-learn/scikit-learn/issues/6196.
        # Overriding __setstate__ to be a noop has the effect of
        # ignoring the pickled __dict__
        pass



"""Observer-based V/Hz control of synchronous motor drives."""

from dataclasses import dataclass, InitVar
import numpy as np
from motulator._helpers import abc2complex
from motulator.control._common import Ctrl, PWM, RateLimiter
from motulator.control.sm._flux_vector import (
    FluxTorqueReference, FluxTorqueReferencePars)
from motulator.control.sm._observers import FluxObserver
from motulator._utils import Bunch


# %%
# pylint: disable=too-many-instance-attributes
@dataclass
class ObserverBasedVHzCtrlPars(FluxTorqueReferencePars):
    """
    Parameters for the control system.

    This class extends FluxTorqueReferencePars with the parameters needed for
    the observer-based V/Hz control.

    Parameters
    ----------
    alpha_psi : float, optional
        Flux control bandwidth (rad/s). The default is 2*pi*50.
    alpha_tau : float
        Torque control bandwidth (rad/s). The default is 2*pi*50.
    alpha_f : float, optional
        Bandwidth of the high-pass filter (rad/s). The default is 2*pi*1.

    """
    alpha_psi: float = 2*np.pi*50
    alpha_tau: InitVar[float] = 2*np.pi*50
    alpha_f: float = 2*np.pi*1

    # pylint: disable=arguments-differ
    def __post_init__(self, par, alpha_tau):
        super().__post_init__(par)
        # Gain k_tau
        G = (par.L_d - par.L_q)/(par.L_d*par.L_q)
        psi_s0 = par.psi_f if par.psi_f > 0 else self.psi_s_min
        if par.psi_f > 0:  # PMSM or PM-SyRM
            c_delta0 = 1.5*par.n_p*(par.psi_f*psi_s0/par.L_d - G*psi_s0**2)
        else:  # SyRM
            c_delta0 = 1.5*par.n_p*G*psi_s0**2
        self.k_tau = alpha_tau/c_delta0


# %%
class ObserverBasedVHzCtrl(Ctrl):
    """
    Observer-based V/Hz control for synchronous motors.

    This observer-based V/Hz control control method is based on [#Tii2022]_.

    Parameters
    ----------
    par : ModelPars
        Machine model parameters.
    ctrl_par : ObserverBasedVHzCtrlPars
        Control system parameters.
    T_s : float, optional
        Sampling period (s). The default is 250e-6.

    Attributes
    ----------
    w_m_ref : callable
        Rotor speed reference (electrical rad/s).

    References
    ----------
    .. [#Tii2022] Tiitinen, Hinkkanen, Kukkola, Routimo, Pellegrino, Harnefors, 
       "Stable and passive observer-based V/Hz control for synchronous Motors," 
       Proc. IEEE ECCE, Detroit, MI, Oct. 2022, 
       https://doi.org/10.1109/ECCE50734.2022.9947858

    """

    # pylint: disable=too-many-instance-attributes, too-many-arguments
    def __init__(self, par, ctrl_par, T_s=250e-6):
        super().__init__()
        self.T_s = T_s
        # Motor parameter estimates
        self.R_s = par.R_s
        self.n_p = par.n_p
        # Controller parameters
        self.alpha_f = ctrl_par.alpha_f
        self.alpha_psi = ctrl_par.alpha_psi
        self.k_tau = ctrl_par.k_tau
        # Subsystems
        self.pwm = PWM()
        self.flux_torque_ref = FluxTorqueReference(ctrl_par)
        self.observer = FluxObserver(par)
        self.rate_limiter = RateLimiter(np.inf)
        self.w_m_ref = callable
        # States
        self.theta_s, self.tau_M_ref = 0, 0

    # pylint: disable=too-many-locals
    def __call__(self, mdl):
        """
        Run the main control loop.

        Parameters
        ----------
        mdl : Drive
            Continuous-time system model for getting the feedback signals.

        Returns
        -------
        T_s : float
            Sampling period (s).
        d_abc : ndarray, shape (3,)
            Duty ratios.

        """
        # Get the speed reference
        if self.rate_limiter is not None:
            w_m_ref = self.rate_limiter(self.T_s, self.w_m_ref(self.clock.t))
        else:
            w_m_ref = self.w_m_ref(self.clock.t)

        # Feedback signals
        i_s_abc = mdl.machine.meas_currents()  # Phase currents
        u_dc = mdl.converter.meas_dc_voltage()  # DC-bus voltage
        u_s = self.pwm.realized_voltage  # Realized voltage from PWM

        # Get the states
        psi_s = self.observer.psi_s
        theta_s = self.theta_s

        # Space vector and coordinate transformation
        i_s = np.exp(-1j*theta_s)*abc2complex(i_s_abc)

        # Get the states
        tau_M_ref = self.tau_M_ref

        # Limited flux and torque references
        psi_s_ref, _ = self.flux_torque_ref(tau_M_ref, w_m_ref, u_dc)

        # Electromagnetic torque
        tau_M = 1.5*self.n_p*np.imag(i_s*np.conj(psi_s))

        # Dynamic frequency
        w_s = w_m_ref - self.k_tau*(tau_M - tau_M_ref)

        # Voltage reference
        err = psi_s_ref - psi_s
        u_s_ref = self.R_s*i_s + 1j*w_s*psi_s_ref + self.alpha_psi*err

        # Data logging
        data = Bunch(
            i_s=i_s,
            psi_s=psi_s,
            psi_s_ref=psi_s_ref,
            t=self.clock.t,
            theta_s=theta_s,
            u_dc=u_dc,
            u_s=u_s,
            w_m_ref=w_m_ref,
            w_s=w_s,
            tau_M=tau_M,
        )
        self.save(data)

        # Update the states
        self.observer.update(self.T_s, u_s, i_s, w_s)
        self.tau_M_ref += self.T_s*self.alpha_f*(tau_M - self.tau_M_ref)
        self.theta_s += self.T_s*w_s  # Next line: limit into [-pi, pi)
        self.theta_s = np.mod(self.theta_s + np.pi, 2*np.pi) - np.pi
        self.clock.update(self.T_s)

        # PWM output
        d_abc = self.pwm(self.T_s, u_s_ref, u_dc, theta_s, w_s)

        return self.T_s, d_abc



"""
*motulator*: Motor Drive Simulator in Python

This software includes continuous-time simulation models for induction machines 
and synchronous machines. Furthermore, selected examples of discrete-time 
control algortihms are also included as well as various utilities.

"""

from motulator._helpers import (
    abc2complex, complex2abc, BaseValues, Sequence, Step)

from motulator._plots import plot, plot_extra

from motulator import control
from motulator import model

__all__ = [
    "abc2complex",
    "complex2abc",
    "BaseValues",
    "Sequence",
    "Step",
    "control",
    "model",
    "plot",
    "plot_extra",
]



"""Common control functions and classes."""

import numpy as np
from motulator._helpers import abc2complex, complex2abc
from motulator._utils import Bunch


# %%
class PWM:
    """
    Duty ratios and realized voltage for three-phase PWM.

    This contains the computation of the duty ratios and the realized voltage. 
    The realized voltage is computed based on the measured DC-bus voltage and 
    the duty ratios. The digital delay effects are taken into account in the 
    realized voltage, assuming the delay of a single sampling period. The total
    delay is 1.5 sampling periods due to the PWM (or ZOH) delay [#Bae2003]_.

    Parameters
    ----------
    six_step: bool, optional
        Enable six-step operation in overmodulation. The default is False.
    
    Attributes
    ----------
    realized_voltage : complex
        Realized voltage (V) in synchronous coordinates.

    References
    ----------
    .. [#Bae2003] Bae, Sul, "A compensation method for time delay of 
       full-digital synchronous frame current regulator of PWM AC drives," IEEE
       Trans. Ind. Appl., 2003, https://doi.org/10.1109/TIA.2003.810660
    
    """

    def __init__(self, six_step=False):
        self.six_step = six_step
        self.realized_voltage = 0
        self._u_ref_lim_old = 0

    @staticmethod
    def six_step_overmodulation(u_s_ref, u_dc):
        """
        Overmodulation up to six-step operation.

        This method modifies the angle of the voltage reference vector in the
        overmodulation region such that the six-step operation is reached 
        [#Bol1997]_.

        Parameters
        ----------
        u_s_ref : complex
            Reference voltage (V) in stator coordinates.
        u_dc : float
            DC-bus voltage (V).

        Returns
        -------
        u_s_ref_mod : complex
            Reference voltage (V) in stator coordinates.

        References
        ----------
        .. [#Bol1997] Bolognani, Zigliotto, "Novel digital continuous control of 
           SVM inverters in the overmodulation range," IEEE Trans. Ind. Appl.,
           1997, https://doi.org/10.1109/28.568019

        """
        # Limited magnitude
        r = np.min([np.abs(u_s_ref), 2/3*u_dc])

        if np.sqrt(3)*r > u_dc:
            # Angle and sector of the reference vector
            theta = np.angle(u_s_ref)
            sector = np.floor(3*theta/np.pi)

            # Angle reduced to the first sector (at which sector == 0)
            theta0 = theta - sector*np.pi/3

            # Intersection angle, see Eq. (9)
            alpha_g = np.pi/6 - np.arccos(u_dc/(np.sqrt(3)*r))

            # Modify the angle according to Eq. (4)
            if alpha_g <= theta0 <= np.pi/6:
                theta0 = alpha_g
            elif np.pi/6 <= theta0 <= np.pi/3 - alpha_g:
                theta0 = np.pi/3 - alpha_g

            # Modified reference voltage
            u_s_ref_mod = r*np.exp(1j*(theta0 + sector*np.pi/3))
        else:
            u_s_ref_mod = u_s_ref

        return u_s_ref_mod

    @staticmethod
    def duty_ratios(u_s_ref, u_dc):
        """
        Compute the duty ratios for three-phase PWM.

        This computes the duty ratios using a symmetrical suboscillation
        method. This method is identical to the standard space-vector PWM.

        Parameters
        ----------
        u_s_ref : complex
            Voltage reference in stator coordinates (V).
        u_dc : float
            DC-bus voltage (V).

        Returns
        -------
        d_abc : ndarray, shape (3,)
            Duty ratios.

        """
        # Phase voltages without the zero-sequence voltage
        u_abc = complex2abc(u_s_ref)

        # Symmetrization by adding the zero-sequence voltage
        u_0 = .5*(np.amax(u_abc) + np.amin(u_abc))
        u_abc -= u_0

        # Preventing overmodulation by means of a minimum phase error method
        m = (2./u_dc)*np.amax(u_abc)
        if m > 1:
            u_abc = u_abc/m

        # Duty ratios
        d_abc = .5 + u_abc/u_dc

        return d_abc

    # pylint: disable=too-many-arguments
    def __call__(self, T_s, u_ref, u_dc, theta, w):
        """
        Compute the duty ratios and update the state.

        Parameters
        ----------
        T_s : float
            Sampling period (s).
        u_ref : complex
            Voltage reference in synchronous coordinates (V).
        u_dc : float
            DC-bus voltage.
        theta : float
            Angle of synchronous coordinates (rad).
        w : float
            Angular speed of synchronous coordinates (rad/s).

        Returns
        -------
        d_abc : ndarray, shape (3,)
            Duty ratios for the next sampling period.

        """
        # Advance the angle due to the computational delay (T_s) and the ZOH
        # (PWM) delay (0.5*T_s)
        theta_comp = theta + 1.5*T_s*w

        # Voltage reference in stator coordinates
        u_s_ref = np.exp(1j*theta_comp)*u_ref

        # Modify angle in the overmodulation region
        if self.six_step:
            u_s_ref = self.six_step_overmodulation(u_s_ref, u_dc)

        # Duty ratios
        d_abc = self.duty_ratios(u_s_ref, u_dc)

        # Realizable voltage
        u_s_ref_lim = abc2complex(d_abc)*u_dc
        u_ref_lim = np.exp(-1j*theta_comp)*u_s_ref_lim

        # Update the voltage estimate for the next sampling instant.
        self.realized_voltage = .5*(self._u_ref_lim_old + u_ref_lim)
        self._u_ref_lim_old = u_ref_lim

        return d_abc


# %%
class PICtrl:
    """
    2DOF PI controller.

    This implements a discrete-time 2DOF PI controller, whose continuous-time
    counterpart is::

        u = k_t*y_ref - k_p*y + (k_i/s)*(y_ref - y)

    where `u` is the controller output, `y_ref` is the reference signal, `y` is
    the feedback signal, and `1/s` refers to integration. The standard PI
    controller is obtained by choosing ``k_t = k_p``. The integrator anti-windup 
    is implemented based on the realized controller output.

    Notes
    -----
    This contoller can be used, e.g., as a speed controller. In this case, `y` 
    corresponds to the rotor angular speed `w_M` and `u` to the torque reference 
    `tau_M_ref`.

    Parameters
    ----------
    k_p : float
        Proportional gain.
    k_i : float
        Integral gain.
    k_t : float, optional
        Reference-feedforward gain. The default is `k_p`.
    u_max : float, optional
        Maximum controller output. The default is inf.

    Attributes
    ----------
    v : float
        Input disturbance estimate.
    u_i : float
        Integral state.

    """

    def __init__(self, k_p, k_i, k_t=None, u_max=np.inf):
        self.u_max = u_max
        # Gains
        self.k_p = k_p
        self.k_t = k_t if k_t is not None else k_p
        self.alpha_i = k_i/self.k_t  # Inverse of the integration time T_i
        # States
        self.v, self.u_i = 0, 0

    def output(self, y_ref, y):
        """
        Compute the controller output.

        Parameters
        ----------
        y_ref : float
            Reference signal.
        y : float
            Feedback signal.

        Returns
        -------
        u : float
            Controller output.

        """
        # Estimate of a disturbance input
        self.v = self.u_i - (self.k_p - self.k_t)*y

        # Controller output
        u = self.k_t*(y_ref - y) + self.v

        # Limit the controller output
        u = np.clip(u, -self.u_max, self.u_max)

        return u

    def update(self, T_s, u_lim):
        """
        Update the integral state.

        Parameters
        ----------
        T_s : float
            Sampling period (s).
        u_lim : float
            Realized (limited) controller output. If the actuator does not
            saturate, ``u_lim = u``.

        """
        self.u_i += T_s*self.alpha_i*(u_lim - self.v)


# %%
class SpeedCtrl(PICtrl):
    """
    2DOF PI speed controller.

    This provides an interface for a speed controller. The gains are initialized
    based on the desired closed-loop bandwidth and the rotor inertia estimate.

    Parameters
    ----------
    J : float
        Total inertia of the rotor (kgm²).
    alpha_s : float
        Closed-loop bandwidth (rad/s). 
    tau_M_max : float, optional
        Maximum motor torque (Nm). The default is inf.

    """

    def __init__(self, J, alpha_s, tau_M_max=np.inf):
        k_p = 2*alpha_s*J
        k_i = alpha_s**2*J
        k_t = alpha_s*J
        super().__init__(k_p, k_i, k_t, tau_M_max)


# %%
class ComplexPICtrl:
    """
    2DOF synchronous-frame complex-vector PI controller.

    This implements a discrete-time 2DOF synchronous-frame complex-vector PI
    controller, whose continuous-time counterpart is [#Bri2000]_::

        u = k_t*i_ref - k_p*i + (k_i + 1j*w*k_t)/s*(i_ref - i)

    where `u` is the controller output, `i_ref` is the reference signal, `i` is
    the feedback signal, `w` is the angular speed of synchronous coordinates, 
    and `1/s` refers to integration. This algorithm is compatible with both real 
    and complex signals. The 1DOF version is obtained by setting ``k_t = k_p``. 
    The integrator anti-windup is implemented based on the realized controller 
    output.

    Parameters
    ----------
    k_p : float
        Proportional gain.
    k_i : float
        Integral gain.
    k_t : float, optional
        Reference-feedforward gain. The default is `k_p`.

    Attributes
    ----------
    v : complex
        Input disturbance estimate.
    u_i : complex
        Integral state.

    Notes
    -----
    This contoller can be used, e.g., as a current controller. In this case, `i`
    corresponds to the stator current and `u` to the stator voltage.
    
    References
    ----------
    .. [#Bri2000] Briz, Degner, Lorenz, "Analysis and design of current 
       regulators using complex vectors," IEEE Trans. Ind. Appl., 2000,
       https://doi.org/10.1109/28.845057

    """

    def __init__(self, k_p, k_i, k_t=None):
        # Gains
        self.k_p = k_p
        self.k_t = k_t if k_t is not None else k_p
        self.alpha_i = k_i/self.k_t  # Inverse of the integration time T_i
        # States
        self.v, self.u_i = 0, 0

    def output(self, i_ref, i):
        """
        Compute the controller output.

        Parameters
        ----------
        i_ref : complex
            Reference signal.
        i : complex
            Feedback signal.

        Returns
        -------
        u : complex
            Controller output.

        """
        # Disturbance input estimate
        self.v = self.u_i - (self.k_p - self.k_t)*i

        # Controller output
        u = self.k_t*(i_ref - i) + self.v

        return u

    def update(self, T_s, u_lim, w):
        """
        Update the integral state.

        Parameters
        ----------
        T_s : float
            Sampling period (s).
        u_lim : complex
            Realized (limited) controller output. If the actuator does not
            saturate, ``u_lim = u``.
        w : float
            Angular speed of the reference frame (rad/s). 

        """
        self.u_i += T_s*(self.alpha_i + 1j*w)*(u_lim - self.v)


# %%
class RateLimiter:
    """
    Rate limiter.

    Parameters
    ----------
    rate_limit : float, optional
        Rate limit. The default is inf.

    """

    def __init__(self, rate_limit=np.inf):
        self.rate_limit = rate_limit
        self._y_old = 0

    def __call__(self, T_s, u):
        """
        Limit the input signal.

        Parameters
        ----------
        u : float
            Input signal.

        Returns
        -------
        T_s : float
            Sampling period (s).
        y : float
            Rate-limited output signal.

        Notes
        -----
        In this implementation, the falling rate limit equals the (negative)
        rising rate limit. If needed, these limits can be separated with minor
        modifications in the class.

        """
        rate = (u - self._y_old)/T_s

        if rate > self.rate_limit:
            # Limit rising rate
            y = self._y_old + T_s*self.rate_limit
        elif rate < -self.rate_limit:
            # Limit falling rate
            y = self._y_old - T_s*self.rate_limit
        else:
            y = u

        # Store the limited output
        self._y_old = y

        return y


# %%
class Clock:
    """Digital clock."""

    def __init__(self):
        self.t = 0
        self.t_reset = 1e9

    def update(self, T_s):
        """
        Update the digital clock.

        Parameters
        ----------
        T_s : float
            Sampling period (s).

        """
        self.t += T_s if self.t < self.t_reset else 0


# %%
class Ctrl:
    """Base class for the control system."""

    def __init__(self):
        self.data = Bunch()  # Data store
        self.clock = Clock()  # Digital clock

    def __call__(self, mdl):
        """
        Run the main control loop.

        The main control loop is callable that returns the sampling
        period `T_s` (float)  and the duty ratios `d_abc` (ndarray, shape (3,)) 
        for the next sampling period.

        Parameters
        ----------
        mdl : Model
            System model containing methods for getting the feedback signals.

        """
        raise NotImplementedError

    def save(self, data):
        """
        Save the internal date of the control system.

        Parameters
        ----------
        data : bunch or dict
            Contains the data to be saved.

        """
        for key, value in data.items():
            self.data.setdefault(key, []).extend([value])

    def post_process(self):
        """
        Transform the lists to the ndarray format.

        This method can be run after the simulation has been completed in order 
        to simplify plotting and analysis of the stored data.

        """
        for key in self.data:
            self.data[key] = np.asarray(self.data[key])



"""This package contains example controllers for induction machines."""

from motulator.control.im._vector import (
    CurrentCtrl,
    CurrentReferencePars,
    CurrentReference,
    ModelPars,
    VectorCtrl,
)

from motulator.control.im._observers import (Observer, FluxObserver)

from motulator.control.im._obs_vhz import (
    ObserverBasedVHzCtrl, ObserverBasedVHzCtrlPars)

from motulator.control.im._vhz import VHzCtrl

__all__ = [
    "CurrentCtrl", "CurrentReferencePars", "CurrentReference", "ModelPars",
    "VectorCtrl", "Observer", "FluxObserver", "ObserverBasedVHzCtrl",
    "ObserverBasedVHzCtrlPars", "VHzCtrl"
]
